import locale               # Python Locale - https://docs.python.org/3/library/locale.html
import pathlib              # Python Pathlib - https://docs.python.org/3/library/pathlib.html
import sys                  # Python Sys - https://docs.python.org/3/library/sys.html

import wx                   # wxPython - https://docs.wxpython.org/

# we don't use XRC, but it's part of the wxFormBuilder template.
# For some reason, late import causes Haiku problems. So import nice and early and get it over with.
import wx.xrc               # wxPython XRC - https://docs.wxpython.org/wx.xrc.1moduleindex.html

from commit.includes.commitapp import CommitApp     # local - override app (for Mac file handling)
from commit.ui.mainwin import MainWin               # local - main window
from commit.includes.customobjects import SplashScreen  # local - custom splash screen implementation

if wx.Platform == "__WXMSW__":
    import ctypes

# Launch the application
if __name__ == '__main__':
    # prevent pointless logging; comment this out to deal with bugs on Py2Exe :-)
    sys.stderr = sys.stdout

    # necessary adjustments in Windows, which has some...unique...behaviors.
    if wx.Platform == "__WXMSW__":
        # wxWidgets is adding Windows dark mode support, but it's not enabled in wxPython yet.
        # When it is ready, uncomment this next line and pull all the custom junk out.
        # wx.SystemOptions.SetOption("msw.dark-mode", 1)

        # tell Windows not to mess with the toolbar colors - https://docs.wxpython.org/wx.ToolBar.html
        wx.SystemOptions.SetOption("msw.remap", 2)

        # tell Windows that we are cool with high-dpi displays (i.e., make things happy and crisp)
        try:
            ctypes.windll.shcore.SetProcessDpiAwareness(True)
        except:  # NOQA; we really want to just continue if this fails for any reason
            pass
    elif wx.Platform == "__WXMAC__":
        wx.SystemOptions.SetOption("osx.openfiledialog.always-show-types", 1)

    app = CommitApp(
        useBestVisual=True
    )

    # init locale
    try:
        locale.setlocale(
            locale.LC_ALL,
            locale.getdefaultlocale()[0]
        )
    except:  # noqa -- we really want to just continue if locales fail
        pass

    # init translations
    _ = app.constants.translation.gettext

    # some product info to the console
    print(app.constants.app_display_name + ": " + app.constants.app_tagline)
    print(app.constants.vendor_display_name + " - " + app.constants.vendor_product_code)

    # handle command line arguments
    for this_arg in sys.argv:
        if this_arg == "--version":
            print("\n" + _("Version") + " " + app.constants.app_version)
            quit()
        elif this_arg == "--reset":
            app.config.Destroy()
            app.config.file_path.unlink(missing_ok=True)
            print("\n" + _("Commit settings have been reset."))
            quit()
        elif this_arg == "--help":
            print("\ncommit: commit [--version|--reset|--help|--debug] [FILE]\n")
            print("    Options:")
            print("      --version output the version number")
            print("      --reset   reset all Commit settings")
            print("      --help    output this help information")
            print("      --debug   show the debugging tools")
            quit()
        elif this_arg == "--debug":
            app.constants.debug = True

    # we got past the special ones; if there are args, we treat the last as a file to open
    if len(sys.argv) > 1:
        app.file_on_open = pathlib.Path(sys.argv[-1])

    # create the main window (will show automatically)
    main_window = MainWin(None, app.file_on_open)

    # show the splash screen, unless suppressed by settings
    if app.config.ReadBool("startup/show_splash"):
        splash = SplashScreen(main_window)
    else:
        main_window.Show()

    if app.constants.debug:
        app.enable_debug_mode(
            wxinsp=True,
            st=True,
            fh=True
        )

    # apploopy
    app.MainLoop()
