import pathlib        # Python Pathlib - https://docs.python.org/3/library/pathlib.html

import appdirs      # AppDirs - https://github.com/ActiveState/appdirs
import wx           # wxPython - https://wxpython.org/


class AppConfig(wx.FileConfig):
    def __init__(self):
        app = wx.GetApp()

        # get the correct directory for saving settings
        self.path = pathlib.Path(
            appdirs.user_config_dir(
                app.constants.app_name,
                app.constants.vendor_name
            )
        )

        '''
        Example config directories:
        
        Windows: ~/AppData/Local/intersanity/commit/
        Mac: ~/Library/Application Support/commit/
        Linux: ~/.config/commit/
        Haiku /boot/home/config/settings/commit/
        '''

        # make sure the directories exist
        self.path.mkdir(parents=True, exist_ok=True)

        # save the file location
        self.file_path = pathlib.Path(
            str(self.path.resolve()) + "/config.ini"
        )

        # call parent init
        super().__init__(
            localFilename=str(self.file_path.resolve()),
            style=wx.CONFIG_USE_LOCAL_FILE
        )

        # set to record back our default values
        self.SetRecordDefaults(True)

        # init startup options
        self.ReadBool("startup/show_splash", True)
        self.ReadBool("startup/open_last_edited", True)

        # init general options
        self.ReadBool("options/auto-save-commit", True)
        self.ReadBool("options/commit-sounds", False)
        self.ReadBool("options/disable-scrolling", False)

        # we want 12pt, but platforms are a mess of differing ratios from pt to px....
        base_default_font_size = 12

        if self.HasEntry("view/font_size"):
            # we have an old font size setting (1.0.0b1 or earlier); need to convert
            default_font_size = self.ReadInt("view/font_size", base_default_font_size)
            self.DeleteEntry("view/font_size")
        elif app.platform == "mac":
            default_font_size = round(1.2 * base_default_font_size)
        elif app.platform == "hku":
            default_font_size = round(0.9 * base_default_font_size)
        else:
            default_font_size = base_default_font_size

        # init the editor font
        default_font = wx.Font(
            wx.FontInfo(
                default_font_size
            ).Family(
                wx.FONTFAMILY_TELETYPE
            )
        )
        self.Read("view/font", default_font.GetNativeFontInfoDesc())

        # init other view options
        self.ReadInt("view/theme", 0)
        self.ReadInt("view/fs_theme", -1)
        self.ReadInt("view/fs_ratio", 4)
        self.ReadBool("view/gen_colors", False)
        self.ReadBool("view/gen_icons", False)

        # init statusbar options
        self.ReadBool("statusbar/show", True)
        self.ReadBool("statusbar/show_word_count", True)
        self.ReadBool("statusbar/show_session_word_count", True)

        # init update values
        self.ReadInt("updates/interval", 2)
        self.ReadInt("updates/last_check", 0)

        # init window positions
        self.ReadInt(
            "window/size_x",
            round(wx.DisplaySize()[1] * 0.9 * 1.3333)
        )
        self.ReadInt(
            "window/size_y",
            round(wx.DisplaySize()[1] * 0.9)
        )
        self.ReadInt("window/pos_x", -1)
        self.ReadInt("window/pos_y", -1)
        self.ReadBool("window/is_maximized", False)
        self.ReadBool("window/is_fullscreen", False)

        # init recent documents
        self.Read("files/current", "")
        self.Read("files/recent", "")

        # init other params
        self.ReadBool("params/pandoc_check", True)

        # save out
        self.Flush()
