import faulthandler         # Python Fault Handling - https://docs.python.org/3/library/faulthandler.html
import pathlib              # Python Pathlib - https://docs.python.org/3/library/pathlib.html
import os                   # Python Misc. OS Interfaces - https://docs.python.org/3/library/os.html
import platform             # Python Platform - https://docs.python.org/3/library/platform.html

import wx                   # wxPython - https://wxpython.org/
import wx.html              # wxPython HTML - https://docs.wxpython.org/wx.html.1moduleindex.html
import wx.lib.inspection    # wxPython Inspection Tool - https://docs.wxpython.org/wx.lib.inspection.html

from codecs import decode as d
from commit import constants                                    # local - app constants
from commit.config import AppConfig                             # local - app config (preferences)
from commit.includes.appupdater import AppUpdater               # local - app updater mechanism
from commit.includes.customobjects import ThemeChangeDetect     # local - theme change detection (for Win)
from commit.includes.recentlist import RecentList               # local - recent list maintainer
from commit.ui.mainwin import MainWin                           # local - mainwin (for class compare)

if wx.Platform == "__WXMSW__":
    import ctypes
elif wx.Platform == "__WXGTK__":
    import gi  # noqa
    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk, GLib  # noqa


class CommitApp(wx.App):
    def __init__(self, *args, **kwargs):
        self.file_on_open = pathlib.Path()
        super().__init__(*args, **kwargs)

        # set our app constants
        # set important app and vendor attributes from constant values
        self.SetAppName(constants.app_name)
        self.SetAppDisplayName(constants.app_display_name)
        self.SetVendorName(constants.vendor_name)
        self.SetVendorDisplayName(constants.vendor_display_name)

        # save our constants to the app object for use elsewhere
        self.constants = constants

        # app config reset flag (so we don't accidentally re-save settings on reset)
        self.is_config_reset = False

        # set platform
        self.platform = "unknown"

        if wx.Platform == "__WXMSW__":
            # wxMSW is Windows
            self.platform = "win"
        elif wx.Platform == "__WXMAC__":
            # wxMAC is MacOS
            self.platform = "mac"
        elif wx.Platform == "__WXGTK__":
            # wxGTK is a multiplatform system
            if platform.system() == "Haiku":
                # Detect if we're in Haiku OS
                self.platform = "hku"
            else:
                # Otherwise, assume GTK means Linux
                self.platform = "lin"

        # set/update the constants that have to be done at runtime

        # modify the product code depending on OS
        if self.platform == "win":  # Windows (MSW)
            this_replace_string = "A"
        elif self.platform == "mac":  # MacOS (Cocoa)
            this_replace_string = "B"
        elif self.platform == "lin":  # Linux (GTK)
            this_replace_string = "C"
        elif self.platform == "hku":  # Haiku (GTK)
            this_replace_string = "D"
        else:  # in case it gets ported elsewhere; can add more OSs later
            this_replace_string = "X"
        self.constants.vendor_product_code = self.constants.vendor_product_code.replace(
            "$", this_replace_string
        )

        # set up the app icon
        self.constants.icon = wx.IconBundle()
        self.constants.icon.AddIcon("commit/assets/icons/appicon16.png", wx.BITMAP_TYPE_PNG)
        self.constants.icon.AddIcon("commit/assets/icons/appicon24.png", wx.BITMAP_TYPE_PNG)
        self.constants.icon.AddIcon("commit/assets/icons/appicon32.png", wx.BITMAP_TYPE_PNG)
        self.constants.icon.AddIcon("commit/assets/icons/appicon48.png", wx.BITMAP_TYPE_PNG)
        self.constants.icon.AddIcon("commit/assets/icons/appicon64.png", wx.BITMAP_TYPE_PNG)
        self.constants.icon.AddIcon("commit/assets/icons/appicon96.png", wx.BITMAP_TYPE_PNG)
        self.constants.icon.AddIcon("commit/assets/icons/appicon128.png", wx.BITMAP_TYPE_PNG)
        self.constants.icon.AddIcon("commit/assets/icons/appicon256.png", wx.BITMAP_TYPE_PNG)
        self.constants.icon.AddIcon("commit/assets/icons/appicon512.png", wx.BITMAP_TYPE_PNG)
        self.constants.icon.AddIcon("commit/assets/icons/appicon1024.png", wx.BITMAP_TYPE_PNG)

        # app config (preferences) object
        self.config = AppConfig()

        # init the recent files list
        self.recents = RecentList()

        # init the app updater
        self.updater = AppUpdater()

        # init the print system
        self.printer = wx.html.HtmlEasyPrinting(
            self.constants.app_display_name
        )

        # determine if the app is in dark mode
        self.is_dark = self.check_is_dark()
        self.is_really_dark = self.check_is_really_dark()
        self.theme = 0
        self.fs_theme = -1
        self.using_gen_icons = False
        self.theme_change_detect = ThemeChangeDetect(self)
        self.Bind(wx.EVT_TIMER, self.fire_theme_change, self.theme_change_detect)

        # handle the Windows dark mode junk
        if self.platform == "win":
            # this has to be done at initial app run (before windows are opened)
            # it will not work reliably as long as the app stays open after a theme setting change
            # but no biggie...all this does is control the right-click context menu colors
            theme_sett = self.config.ReadInt("view/theme")

            # 0 = default; 1 = automatic; 2 = force dark; 3 = force light
            if theme_sett == 1:
                win_sett = 3
            elif theme_sett == 2:
                win_sett = 2
            else:  # 0 or no valid theme setting - follow system
                win_sett = 1
            ctypes.windll.uxtheme[135](  # SetPreferredAppMode (uxtheme@135)
                ctypes.c_int(
                    win_sett
                )
            )

    def get_api_str(self):
        return d(self.constants.api_str, "hex").decode()

    def check_is_dark(self, bg_color=False, fg_color=False, fs_override=False):
        if self.platform == "win" and not wx.CheckOsVersion(10, 0, 19041):
            # if we have an older version of Windows, NOPE.
            return False

        if not bg_color and not fg_color:
            if fs_override:
                theme_sett = self.constants.themes[
                    self.config.ReadInt("view/fs_theme")
                ][0]
            else:
                theme_sett = self.constants.themes[
                    self.config.ReadInt("view/theme")
                ][0]

            if theme_sett == "light":
                return False
            elif theme_sett == "dark":
                return True

            # Freakin' flatpak :-/ the normal dark detection doesn't work. We have to do it a hacky, stupid way
            # Basically...does the theme say it's dark in the name?
            if os.getenv("container") == "flatpak":
                flat_dark = False
                gtk_theme_name = Gtk.Settings.get_default().get_property("gtk-theme-name")
                for this_flag in self.constants.dark_flags:
                    if this_flag in gtk_theme_name:
                        flat_dark = True
                return flat_dark

            # if we don't have defined colors, we want main luminance (dark vs. light)
            if (
                hasattr(wx.SystemSettings, "GetAppearance") and
                callable(wx.SystemSettings.GetAppearance)
            ):
                # if we have a new enough wxPython to just ask if we're dark, do
                return wx.SystemSettings.GetAppearance().IsDark()

            # okay, we can't just ask wxPython... too old. Sigh.
            # let's set our window colors and go on with the luminance comparison
            bg_color = wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW)
            fg_color = wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOWTEXT)

        # we need to compare the colors manually, either because that's what was requested
        # OR because we have to manually check the window luminance
        if (
            hasattr(bg_color, "GetLuminance") and
            callable(bg_color.GetLuminance)
        ):
            # if this version of wxPython has GetLuminance, just use those and subtract
            lume_diff = fg_color.GetLuminance() - bg_color.GetLuminance()
        else:
            # we don't have that function either, calculate manually -- eyeroll -- see wxWidgets colourcmn.cpp
            def comp_lumi(color):
                return (
                    (0.299 * color.Red()) +
                    (0.587 * color.Green()) +
                    (0.114 * color.Blue())
                ) / 255
            lume_diff = comp_lumi(fg_color) - comp_lumi(bg_color)

        # computed!
        return lume_diff > 0.2

    # REALLY check if dark
    def check_is_really_dark(self):
        return self.check_is_dark(
            wx.SystemSettings.GetColour(wx.SYS_COLOUR_FRAMEBK),
            wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOWTEXT)
        )

    # enable the debug mode
    def enable_debug_mode(self, fh=False, wxinsp=False, st=False):
        if fh:
            faulthandler.enable()

        if wxinsp:
            wx.lib.inspection.InspectionTool().Show()

        if st:
            self.RedirectStdio()
            _ = self.constants._  # noqa
            print(_("stdio and stderr output will be posted here."))

    # close the entire application by looping through the windows
    def force_close_app(self):
        self.theme_change_detect.Stop()
        for this_win in wx.GetTopLevelWindows():  # noqa
            this_win.Close(True)

    @staticmethod
    def fire_theme_change(event=None):
        event.Skip()

        for this_win in wx.GetTopLevelWindows():  # noqa
            if type(this_win) == MainWin:
                this_win.do_theme_change()

    # handle Mac-specific stuff: double-clicking ctxt files and drag-drop, etc.
    def MacOpenFiles(self, filenames):
        main_win = self.GetMainTopWindow()
        file_requested = pathlib.Path(filenames[0])

        if file_requested.is_file() and main_win is None:
            self.file_on_open = file_requested
        else:
            main_win.file.open(filenames[0])
