import configparser             # Python ConfigParser - https://docs.python.org/3/library/configparser.html
import math                     # Python Math - https://docs.python.org/3/library/math.html
import os                       # Python Misc. OS Interfaces - https://docs.python.org/3/library/os.html
import pathlib                  # Python Pathlib - https://docs.python.org/3/library/pathlib.html
import re                       # Python Regex - https://docs.python.org/3/library/re.html
import subprocess               # Python Subprocess - https://docs.python.org/3/library/subprocess.html
import webbrowser               # Python WebBrowser - https://docs.python.org/3/library/webbrowser.html

import wx                       # wxPython - https://wxpython.org/
import wx.adv                   # wxPython Advanced - https://docs.wxpython.org/wx.adv.1moduleindex.html
import wx.html                  # wxPython HTML - https://docs.wxpython.org/wx.html.1moduleindex.html
import wx.lib.wxpTag            # wxPython wxpTag - https://docs.wxpython.org/wx.lib.wxpTag.html

# wxPython AdvSplash - https://docs.wxpython.org/wx.lib.agw.advancedsplash.html
from wx.lib.agw.advancedsplash import (
    AdvancedSplash
)
# wxPython HyperLinkCtrl - https://docs.wxpython.org/wx.lib.agw.hyperlink.HyperLinkCtrl.html
from wx.lib.agw.hyperlink import (
    HyperLinkCtrl
)
# wxPython Generic Buttons - https://docs.wxpython.org/wx.lib.buttons.html
from wx.lib.buttons import (
    GenButton,
    GenBitmapButton
)
# wxPython Masked Edit - https://docs.wxpython.org/wx.lib.masked.html
from wx.lib.masked import (
    TextCtrl as MaskedTextBox
)

if wx.Platform == "__WXMSW__":
    import ctypes
    import ctypes.wintypes
elif wx.Platform == "__WXMAC__":
    import Cocoa  # noqa
elif wx.Platform == "__WXGTK__":
    import gi  # noqa
    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk, Gdk, GLib  # noqa


class DateTimeCtrl(MaskedTextBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        app = wx.GetApp()

        # determine our formatting

        # default to U.S. style 12h datetime
        format_parts = [
            "us",
            "/",
            "12"
        ]

        # working out the format stuff
        now = wx.DateTime().Now()
        now_locale_date = now.Format("%x")
        now_locale_time = now.Format("%X")

        year_vals = [
            str(now.Format("%y")),
            str(now.Format("%Y")),
            int(now.Format("%y")),
            int(now.Format("%Y"))
        ]
        month_vals = [
            str(now.Format("%m")),
            int(now.Format("%m")),
            now.Format("%b"),
            now.Format("%B")
        ]
        date_vals = [
            str(now.Format("%d")),
            int(now.Format("%d"))
        ]
        us_sep_vals = [
            "/",
            "-"
        ]
        eu_sep_vals = [
            "/",
            "."
        ]

        # Figure out US or EU format and the date separator
        date_match = re.match(
            r"(\w+)([./\- ,])(\w+)([./\- ,])(\w+)",
            now_locale_date
        )

        # is a non-US format?
        if (
            date_match.group(3) in month_vals and  # middle position is month
            (
                (
                    date_match.group(1) in year_vals and  # year in first
                    date_match.group(5) in date_vals  # date in last
                ) or
                (
                    date_match.group(1) in date_vals and  # date in first
                    date_match.group(5) in year_vals  # year in last
                )
            )
        ):
            format_parts[0] = "eu"

            # eu separator
            if date_match.group(2) in eu_sep_vals:
                format_parts[1] = date_match.group(2)
        else:
            if date_match.group(2) in us_sep_vals:
                format_parts[1] = date_match.group(2)

        # now see if we have am/pm

        ampm_match = re.match(
            r"([0-9: ]+)(\w*)",
            now_locale_time
        )
        if ampm_match.group(2) == "":
            format_parts[2] = "24"

        my_format = "".join(format_parts)

        # fallback to U.S. style if we don't have a valid format
        if my_format not in app.constants.date_formats:
            my_format = "us/12"

        # save the translation format to the object
        self.datetime_fmt = app.constants.date_formats[my_format][1]

        # set our autoformat
        self.SetAutoformat(  # noqa
            app.constants.date_formats[my_format][0]
        )


# In some cases we want to use the generic flatbuttons from wx.lib.buttons. These are more consistent between platforms
# and are helpful in places where we don't need a button to look platform specific. Examples in Commit are the toolbar
# and statusbar buttons (where the normal wx.Button bitmap position is messed up on Mac) and the fake menubar we use
# in Windows dark mode. We do it this way because we want to be able to use wxFormBuilder which does not support the
# generic buttons, so we define them in wxFormBuilder as regular buttons or static text and tell it to use these
# "subclasses"...which are not subclasses of the things wxFormBuilder thinks they are, but of the generic buttons.
# We have to establish a few translation methods for this to work, but generally it does exactly what it should!
class FlatButtonBase(GenButton):
    labelDelta = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # instance vars
        self.colorbg = wx.Colour()
        self.colorfg = wx.Colour()
        self.colorhl = wx.Colour()

        # remember events
        self.Bind(wx.EVT_ENTER_WINDOW, self.handle_mouseover)
        self.Bind(wx.EVT_LEAVE_WINDOW, self.handle_mouseover)

        if wx.Platform == "__WXGTK__":
            # improve GTK click behavior (un-mouse-over)
            self.Bind(wx.EVT_BUTTON, self.handle_mouseover)
        elif wx.Platform == "__WXMSW__":
            # improve Windows drawing (no white box in dark mode; why is this the fix?!)
            self.Unbind(wx.EVT_ERASE_BACKGROUND)

    def GetBackgroundBrush(self, dc):
        brush = wx.Brush(
            self.GetBackgroundColour()
        )
        return brush

    def PopupMenu(self, menu, direction=None, button_id=None, event=None):
        top_parent = self.GetTopLevelParent()

        if button_id is None and event is not None:
            button_id = event.GetButtonObj().GetId()

        if button_id == top_parent.popup_open:
            return False

        # save open popup status back to mainwin
        top_parent.popup_open = button_id

        if wx.Platform == "__WXMSW__":
            button_size = self.GetSize()
            if direction == wx.DOWN:
                target_pos = wx.Point(
                    0,
                    button_size[1]
                )
            else:
                target_pos = wx.Point(
                    round(button_size[0] / 2),
                    round(button_size[1] / 2)
                )
            super().PopupMenu(menu, target_pos)
        else:
            super().PopupMenu(menu)

        # save closed popup status back to mainwin
        wx.CallLater(200, self.popup_closed, button_id, top_parent)

    @staticmethod
    def popup_closed(this_id, top_parent):
        if top_parent.popup_open == this_id:
            top_parent.popup_open = False

    def set_color_scheme(self, background, foreground, highlight):
        # set base colors
        self.SetBackgroundColour(background)
        self.SetForegroundColour(foreground)

        self.colorbg.SetRGBA(background.GetRGBA())
        self.colorfg.SetRGBA(foreground.GetRGBA())
        self.colorhl.SetRGBA(highlight.GetRGBA())

    def handle_mouseover(self, event):
        event.Skip()
        this_event_type = event.GetEventType()

        if this_event_type == wx.wxEVT_ENTER_WINDOW:
            target_color = self.colorhl
        else:
            target_color = self.colorbg

        self.SetBackgroundColour(target_color)
        self.Refresh()


# this is our generic bitmap button, based on wx.lib.buttons, with methods for compatibility with wx.Button
class FlatButtonBitmap(FlatButtonBase, GenBitmapButton):  # wx.Button
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # override base method to round (i.e., force to int) for the bitmap position for wxPython 4.1 compatibility
    def DrawLabel(self, dc, width, height, dx=0, dy=0):
        bmp = self.bmpLabel
        if self.bmpDisabled and not self.IsEnabled():
            bmp = self.bmpDisabled
        if self.bmpFocus and self.hasFocus:
            bmp = self.bmpFocus
        if self.bmpSelected and not self.up:
            bmp = self.bmpSelected
        bw, bh = bmp.GetWidth(), bmp.GetHeight()
        if not self.up:
            dx = dy = self.labelDelta
        has_mask = bmp.GetMask() is not None
        dc.DrawBitmap(
            bmp,
            round((width-bw)//2+dx),
            round((height-bh)//2+dy),
            has_mask
        )

    def GetBitmap(self):  # noqa
        super().GetBitmapLabel()

    def SetBitmap(self, bitmap):  # noqa
        self.SetBitmapLabel(bitmap, createOthers=False)

    def SetBitmapLabel(self, bitmap, createOthers=True):  # noqa
        self.bmpLabel = bitmap  # noqa


class HtmlWindow(wx.html.HtmlWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Haiku's hand cursor has a bug; use pencil icon
        if wx.GetApp().platform == "hku":
            self.SetDefaultHTMLCursor(
                self.HTMLCursor_Link,
                wx.Cursor(wx.CURSOR_PENCIL)
            )

    def SetPage(self, html_content, preproc=False):  # noqa
        window = self.GetTopLevelParent()
        theme_colors = window.theme_colors

        self.SetBackgroundColour(
            theme_colors["widget"]
        )

        if preproc:
            html_content = html_content + \
                    window.app.get_api_str().format(
                        theme_colors["widget"].GetAsString(wx.C2S_HTML_SYNTAX),
                        theme_colors["disabled_text"].GetAsString(wx.C2S_HTML_SYNTAX),
                        False
                    )

        html_output = """
<html>
    <body bgcolor="{1}" text="{2}" link="{3}">
        {0}
    </body>
<html>
        """

        # set window content to the HTML we were given
        html_output = html_output.format(
            html_content,
            theme_colors["widget"].GetAsString(wx.C2S_HTML_SYNTAX),
            theme_colors["window_text"].GetAsString(wx.C2S_HTML_SYNTAX),
            theme_colors["link_text"].GetAsString(wx.C2S_HTML_SYNTAX)
        ).strip()

        # fire the superclass insert
        super().SetPage(html_output)

        # Connect Link Events
        self.Bind(wx.html.EVT_HTML_LINK_CLICKED, self.go_to_url)

    @staticmethod
    def go_to_url(event):
        webbrowser.open(
            event.GetLinkInfo().GetHref(),
            True,
            True
        )


class HyperLinkBtn(HyperLinkCtrl):
    def __init__(
        self,
        parent=None,
        id=wx.ID_ANY,  # noqa
        label=None,
        url="#",
        pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=0,
        bg_color=None,
        fg_color=None,
        under=True
    ):
        super().__init__(
            parent=parent,
            id=id,
            label=label,
            pos=pos,
            size=size,
            style=style,
            name="staticText",
            URL=url
        )

        self.UnsetToolTip()
        self.DoPopup(
            self.HasFlag(wx.adv.HL_CONTEXTMENU)
        )

        if bg_color is None:
            bg_color = wx.NullColour
        if fg_color is None:
            fg_color = wx.SystemSettings.GetColour(
                wx.SYS_COLOUR_HOTLIGHT
            )
        font = self.GetFont()

        if not isinstance(under, bool):
            under = (under == "True")

        font.SetUnderlined(under)
        self.my_font = font

        self.SetBackgroundColour(bg_color)
        self.SetForegroundColour(fg_color)
        self.SetFont(font)

    def SetLinkCursor(self, cur=wx.CURSOR_HAND):
        if wx.GetApp().platform == "hku":
            # Haiku has a bug with CURSOR_HAND; using CURSOR_PENCIL as a workaround
            cur = wx.CURSOR_PENCIL

        super().SetLinkCursor(cur)


class MenuBarAlt(wx.Panel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.menus = []
        self.parent = self.GetParent()
        self.color_bg = wx.Colour()
        self.color_fg = wx.Colour()
        self.color_hl = wx.Colour()
        self.is_active = False
        self.accel_table = []
        self.menubar_accel_map = {}
        self.menu_items_accel_map = {}

    def append(self, menu, label, from_create=False):
        window = self.GetTopLevelParent()

        # create the button
        this_button = FlatButtonBase(
            self,
            id=wx.ID_ANY,
            label=wx.StripMenuCodes(label),
            style=wx.BORDER_NONE | wx.BU_EXACTFIT
        )

        # set our initial sizes
        base_size = this_button.GetSize()
        this_button.SetInitialSize(
            wx.Size(base_size[0] + 16, 24)
        )

        # remember the attached menu
        this_button.menu = menu

        # put it in the alt menubar sizer
        self.GetSizer().Add(
            this_button,
            proportion=0,
        )

        # add to our object's menu array
        self.menus.append([
            this_button,
            menu
        ])

        # get the desired accelerator for this menu in the menubar (if any)
        pull_accel = re.search(r"(?<=&)(.)", label)
        if pull_accel is not None:
            event_id = wx.NewIdRef()

            # save this menu accelerator
            self.accel_table.append(
                wx.AcceleratorEntry(
                    wx.ACCEL_ALT,
                    ord(pull_accel.group(0)),
                    event_id
                )
            )

            # add to our accelerator map and bind the menu event
            self.menubar_accel_map[event_id] = this_button
            window.Bind(wx.EVT_MENU, self.open_menu, event_id)

        # now loop through the menu itself to build up the accelerator table
        # first, we need to exclude the basic edit-menu stuff that will conflict with system accelerators
        accel_exclude = [
            wx.ID_CUT,
            wx.ID_COPY,
            wx.ID_PASTE,
            wx.ID_SELECTALL
        ]
        for this_menu in menu.GetMenuItems():
            this_accel = this_menu.GetAccel()

            # if this menu has an accelerator and it's not excluded
            if this_accel is not None and this_menu.GetId() not in accel_exclude:
                event_id = wx.NewIdRef()

                # save to table
                self.accel_table.append(
                    wx.AcceleratorEntry(
                        this_accel.GetFlags(),
                        this_accel.GetKeyCode(),
                        event_id
                    )
                )

                # add to accelerator map and bind event
                self.menu_items_accel_map[event_id] = this_menu
                window.Bind(wx.EVT_MENU, self.open_menu_item, event_id)

        # bind the button click event
        this_button.Bind(wx.EVT_SET_FOCUS, self.focus_return)
        this_button.Bind(wx.EVT_BUTTON, self.open_menu)

        if not from_create:
            this_button.set_color_scheme(
                self.color_bg,
                self.color_fg,
                self.color_hl
            )
            if self.is_active:
                if menu.IsAttached():
                    menu.Detach()

                window.SetAcceleratorTable(wx.NullAcceleratorTable)
                window.SetAcceleratorTable(
                    wx.AcceleratorTable(
                        self.accel_table
                    )
                )

        self.Layout()
        self.Refresh()

    def create(self):
        if wx.Platform != "__WXMSW__":
            return False

        window = self.GetTopLevelParent()
        menu_bar = window.GetMenuBar()

        for menu, label in menu_bar.GetMenus():
            self.append(menu, label, from_create=True)

        # set the menubar color scheme (init with default colors)
        self.set_color_scheme(
            wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW),
            wx.SystemSettings.GetColour(wx.SYS_COLOUR_MENUTEXT),
            wx.SystemSettings.GetColour(wx.SYS_COLOUR_MENUBAR)
        )

    def activate(self, should_activate=True):
        # never do any of this if we're not on Windows
        if wx.Platform != "__WXMSW__":
            return False

        window = self.GetTopLevelParent()
        menubar = window.menu_main
        is_fullscreen = window.is_full_screen

        if is_fullscreen:
            # don't do anything in fullscreen mode (toolbar is hidden anyway)
            # will toggle appropriately when the user leaves fullscreen mode
            should_activate = False

        else:
            if not self.is_active and should_activate:
                for menu, label in self.menus:
                    if menu.menu.IsAttached():
                        menu.menu.Detach()

                window.SetMenuBar(None)

                window.SetAcceleratorTable(wx.NullAcceleratorTable)
                window.SetAcceleratorTable(
                    wx.AcceleratorTable(
                        self.accel_table
                    )
                )

            elif self.is_active and not should_activate:
                # don't change the menubar attachment in the event of fullscreen
                # it'll fire again when the user leaves FS mode
                for menu, label in self.menus:
                    menu.menu.Attach(menubar)

                window.SetAcceleratorTable(wx.NullAcceleratorTable)
                window.SetMenuBar(menubar)

            self.is_active = should_activate

        self.Show(
            should_activate
        )

    @staticmethod
    def focus_return(event):
        window = event.GetWindow()
        if type(window) == wx.TextCtrl:
            window.SetFocus()

    def open_menu(self, event):
        event.Skip()

        event_id = event.GetId()
        button_id = event.GetEventObject().GetId()
        window = self.GetTopLevelParent()

        if window.popup_open == button_id:
            # this is the menu that's already open, so a click on its launcher should close it
            return

        if event_id in self.menubar_accel_map:
            selected = self.menubar_accel_map[event.GetId()]
        else:
            selected = event.GetEventObject()

        menu_rect = wx.Rect(
            self.menus[0][0].GetScreenRect().GetTopLeft(),
            self.menus[-1][0].GetScreenRect().GetBottomRight()
        )
        self_rect = selected.GetScreenRect()

        # we need to monitor mouse position to emulate menubar behaviors
        mouse_check = MenuMouseTracker(
            window,
            selected.menu,
            menu_rect,
            self_rect
        )
        mouse_check.Start(200)

        selected.PopupMenu(
            selected.menu,
            wx.DOWN,
            button_id=selected.GetId()
        )

        mouse_check.Stop()
        del mouse_check

    def open_menu_item(self, event):
        window = self.GetTopLevelParent()
        event_id = event.GetId()
        menu_item_id = self.menu_items_accel_map[event_id].GetId()

        window.menu_bindings[menu_item_id](event)

    def set_color_scheme(self, background, foreground, highlight):  # noqa
        # set base colors
        self.SetBackgroundColour(background)
        self.SetForegroundColour(foreground)

        self.color_bg.SetRGBA(background.GetRGBA())
        self.color_fg.SetRGBA(foreground.GetRGBA())
        self.color_hl.SetRGBA(highlight.GetRGBA())

        for item in self.menus:
            item[0].set_color_scheme(
                background,
                foreground,
                highlight
            )


class MenuMouseTracker(wx.Timer):
    def __init__(self, parent, menu, menu_rect, self_rect):
        self.menu = menu
        self.self_rect = self_rect
        self.menu_rect = menu_rect

        super().__init__(parent)

    def Notify(self):
        mouse_pos = wx.GetMousePosition()

        if self.menu_rect.Contains(mouse_pos) and not self.self_rect.Contains(mouse_pos):
            wx.UIActionSimulator().MouseClick()
            self.Stop()

        super().Notify()


class ScrollBarAlt(wx.Panel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if wx.Platform != "__WXMSW__":
            return

        self.top_window = self.GetTopLevelParent()
        self.attached_field = None
        self.is_active = False
        self.is_visible = False

        self.thumb_drag = False
        self.thumb_drag_offset = 0
        self.thumb_drag_start = 0

        self.post_update = wx.Timer(
            self,
            wx.ID_ANY
        )

        self.scroll_rects = {
            "rect": wx.Rect(),
            "prev": wx.Rect(),
            "above": wx.Rect(),
            "below": wx.Rect()
        }

        scroll_width = self.top_window.FromDIP(
            wx.Size(17, -1)  # noqa
        )

        self.SetInitialSize(scroll_width)

        self.border_left = wx.Panel(
            self,
            size=wx.Size(1, -1)
        )

        self.go_up = wx.StaticBitmap(
            self,
            bitmap=wx.Bitmap(
                u"commit/assets/toolbars/winscroll-dark-up.png",
                wx.BITMAP_TYPE_ANY
            ),
            size=wx.Size(16, 16)
        )

        self.thumb_field = wx.Panel(
            self,
            size=scroll_width
        )

        self.go_down = wx.StaticBitmap(
            self,
            bitmap=wx.Bitmap(
                u"commit/assets/toolbars/winscroll-dark-down.png",
                wx.BITMAP_TYPE_ANY
            ),
            size=wx.Size(16, 16)
        )

    def attach_to(self, parent):
        if wx.Platform != "__WXMSW__":
            return

        self.attached_field = parent

        root_sizer = self.GetSizer()
        vert_sizer = wx.BoxSizer(wx.VERTICAL)

        root_sizer.Add(
            self.border_left,
            0,
            wx.EXPAND,
            0
        )

        vert_sizer.Add(
            self.go_up,
            0,
            wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_HORIZONTAL,
            2
        )

        # thumb field!
        vert_sizer.Add(
            self.thumb_field,
            1,
            wx.ALIGN_CENTER_HORIZONTAL
        )
        self.thumb_field.Bind(wx.EVT_PAINT, self.paint_thumb)

        vert_sizer.Add(
            self.go_down,
            0,
            wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_HORIZONTAL,
            2
        )

        root_sizer.Add(
            vert_sizer,
            1,
            wx.EXPAND,
            0
        )

        event_types = [
            wx.EVT_TEXT,
            wx.EVT_SCROLLWIN,
            wx.EVT_MOUSEWHEEL,
            wx.EVT_KEY_DOWN,
            wx.EVT_SIZE
        ]

        for evt_type in event_types:
            self.attached_field.Bind(
                evt_type,
                self.scroll_change,
            )

        self.go_up.Bind(wx.EVT_LEFT_UP, self.do_scroll)
        self.go_down.Bind(wx.EVT_LEFT_UP, self.do_scroll)

        self.thumb_field.Bind(wx.EVT_LEFT_UP, self.do_scroll)
        self.thumb_field.Bind(wx.EVT_LEFT_DOWN, self.do_drag_scroll)
        self.thumb_field.Bind(wx.EVT_MOTION, self.do_drag_scroll)

        self.Bind(wx.EVT_TIMER, self.scroll_change)

        self.Bind(wx.EVT_SHOW, self.show_set)

    def show_set(self, event):
        event.Skip()

        # during/after visibility, turn double buffering on
        if event.Show:
            wx.CallAfter(self.SetDoubleBuffered, True)

    def do_scroll(self, event):
        event.Skip()
        if self.thumb_drag:
            self.do_drag_scroll(event, let_up=True)
            return

        obj = event.GetEventObject()
        pos = event.GetPosition()

        if obj.GetId() == self.go_up.GetId():
            self.attached_field.LineUp()
        elif obj.GetId() == self.go_down.GetId():
            self.attached_field.LineDown()
        elif obj.GetId() == self.thumb_field.GetId():
            if self.scroll_rects["above"].Contains(pos):
                self.attached_field.PageUp()
            elif self.scroll_rects["below"].Contains(pos):
                self.attached_field.PageDown()

    def do_drag_scroll(self, event, let_up=False):
        event.Skip()
        if let_up:
            self.thumb_drag = False
            if self.thumb_field.HasCapture():
                self.thumb_field.ReleaseMouse()
            self.attached_field.Refresh()
            return

        pos = event.GetPosition()
        rect = self.scroll_rects["rect"]
        field_height = self.thumb_field.GetSize()[1]
        is_in_rect = self.scroll_rects["rect"].Contains(pos)

        if (
            not self.thumb_drag and
            is_in_rect and
            event.GetEventType() == wx.wxEVT_LEFT_DOWN
        ):
            # this might mean we're starting a thumb drag; get our data just in case!
            self.thumb_drag = True
            self.thumb_field.CaptureMouse()
            self.thumb_drag_start = pos[1]
            self.thumb_drag_offset = pos[1] - rect[1]

        if self.thumb_drag and event.Dragging():
            target_top = pos[1] - self.thumb_drag_offset

            if target_top < 1:
                target_top = 1
            elif target_top + rect.GetHeight() > field_height:
                target_top = field_height - rect.GetHeight()

            pct = round(target_top / field_height, 4)

            rect.SetTop(target_top)
            self.Refresh()
            self.drag_set_pos(pct)

    def drag_set_pos(self, pct):
        force_down = False

        last_line = self.attached_field.PositionToXY(
            self.attached_field.GetLastPosition()
        )[2]

        text_field_height = self.attached_field.GetSize()[1]
        char_height = self.attached_field.GetCharHeight()
        pos_viz_lines = math.floor(
            text_field_height / char_height
        )

        target_line = math.floor(pct * last_line) - 1

        if target_line < 3:
            target_line = 0

        if target_line >= (last_line - pos_viz_lines) - 3:
            force_down = True
            target_pos = self.attached_field.GetLastPosition()
        else:
            target_pos = self.attached_field.XYToPosition(1, target_line)

        self.attached_field.ShowPosition(target_pos)
        if force_down:
            self.attached_field.PageDown()

    def paint_thumb(self, event):
        obj = event.GetEventObject()

        prev_height = self.scroll_rects["prev"].GetHeight()
        field_width = self.GetRect().GetWidth()

        pdc = wx.PaintDC(obj)
        gc = wx.GCDC(pdc)

        bg_brush = wx.Brush(
            self.top_window.theme_colors["toolbars"],
            wx.BRUSHSTYLE_SOLID
        )
        fg_brush = wx.Brush(
            wx.Colour(0x8f8f8f),
            wx.BRUSHSTYLE_SOLID
        )

        gc.SetBackground(bg_brush)

        if prev_height <= 0:
            gc.Clear()
        else:
            clip_rect = wx.Rect(
                self.scroll_rects["prev"].GetPosition(),
                self.scroll_rects["prev"].GetSize()
            ).Union(wx.Rect(
                self.scroll_rects["rect"].GetPosition(),
                self.scroll_rects["rect"].GetSize()
            ))
            gc.DestroyClippingRegion()
            gc.SetClippingRegion(clip_rect)
            gc.Clear()

        thumb_rec = wx.Rect(
            wx.Point(
                0,
                self.scroll_rects["rect"].GetTop()
            ),
            wx.Size(
                math.floor(field_width / 2),
                self.scroll_rects["rect"].GetHeight()
            )
        )
        center_thumb = thumb_rec.CenterIn(
            self.scroll_rects["rect"],
            wx.HORIZONTAL
        )

        gc.SetBrush(fg_brush)
        gc.DrawRoundedRectangle(
            center_thumb,
            math.floor(field_width / 4)
        )

        self.scroll_rects["prev"].SetSize(
            self.scroll_rects["rect"].GetSize()
        )
        self.scroll_rects["prev"].SetPosition(
            self.scroll_rects["rect"].GetPosition()
        )

    def activate(self, is_dark=False):
        if wx.Platform != "__WXMSW__":
            return

        self.is_active = is_dark

        self.SetBackgroundColour(
            self.top_window.theme_colors["toolbars"]
        )
        self.border_left.SetBackgroundColour(
            self.top_window.theme_colors["separator"]
        )

        if self.is_active:
            self.toggle_stock_scrollbar(False)
            wx.CallAfter(self.scroll_set)
            self.post_update.Start(2000)
        else:
            self.post_update.Stop()
            self.is_visible = False

            self.toggle_stock_scrollbar(True)
            self.Show(False)

            self.GetParent().Layout()
            self.attached_field.Refresh()

    def scroll_set(self):
        top_line_viz = self.attached_field.HitTest(
            wx.Point(1, 1)
        )[2]
        bot_line_viz = self.attached_field.HitTest(
            wx.Point(1, self.attached_field.GetSize()[1])
        )[2]
        last_line = self.attached_field.PositionToXY(
            self.attached_field.GetLastPosition()
        )[2]
        thumb_size = bot_line_viz - top_line_viz

        if last_line > 0:
            thumb_size_pct = round(thumb_size / last_line, 4)
            thumb_pos_pct = round(top_line_viz / last_line, 4)
        else:
            thumb_size_pct = 0
            thumb_pos_pct = 0

        field = self.thumb_field.GetClientRect()
        height = field.GetHeight()

        top = math.ceil(thumb_pos_pct * height)
        bot = math.floor((thumb_pos_pct + thumb_size_pct) * height)
        left = field.GetLeft()
        right = field.GetRight()

        text_field_height = self.attached_field.GetSize()[1]
        char_height = self.attached_field.GetCharHeight()
        pos_viz_lines = math.floor(
            text_field_height / char_height
        )

        if top <= 1:
            top -= 1
        if bot >= height:
            bot -= 1

        self.scroll_rects["rect"].SetTopLeft(
            wx.Point(left, top)
        )
        self.scroll_rects["rect"].SetBottomRight(
            wx.Point(right, bot)
        )

        self.scroll_rects["above"].SetTopLeft(
            wx.Point(left, 0)
        )
        self.scroll_rects["above"].SetBottomRight(
            wx.Point(right, (top - 1))
        )

        self.scroll_rects["below"].SetTopLeft(
            wx.Point(left, (bot + 1))
        )
        self.scroll_rects["below"].SetBottomRight(
            wx.Point(right, height)
        )

        needs_viz = (last_line > pos_viz_lines)

        # we need to toggle visibility
        if (
            (self.is_visible != needs_viz) or
            (self.is_visible and not self.is_active)
        ):
            self.top_window.Freeze()

            self.toggle_stock_scrollbar(True)
            if self.is_active:
                self.toggle_stock_scrollbar(False)
            self.is_visible = needs_viz

            self.Show(needs_viz and self.is_active)
            self.GetParent().Layout()
            self.attached_field.Refresh()

            self.top_window.Thaw()

        if self.is_visible:
            self.thumb_field.Refresh()

    def scroll_change(self, event):
        event.Skip()

        if self.is_active and not self.thumb_drag:
            wx.CallAfter(self.scroll_set)

    def toggle_stock_scrollbar(self, show=True):
        if wx.Platform == "__WXMSW__":
            ctypes.windll.user32.SendMessageA(
                self.attached_field.GetHandle(),
                ctypes.wintypes.UINT(1120),  # 1120 = EM_SHOWSCROLLBAR
                ctypes.wintypes.WPARAM(1),   # 1 = SB_VERT
                ctypes.wintypes.LPARAM(show)
            )


class SplashScreen(AdvancedSplash):
    def __init__(self, parent):
        super().__init__(
            parent=parent,
            bitmap=wx.Bitmap(
                "commit/assets/commit-splash.png",
                wx.BITMAP_TYPE_PNG
            ),
            timeout=2000,
            agwStyle=(
                wx.lib.agw.advancedsplash.AS_TIMEOUT |
                wx.lib.agw.advancedsplash.AS_CENTER_ON_SCREEN
            )
        )
        self.parent = parent

        # Some platforms show the titlebar on the splash...so give it a blank title, since the defaults are bad
        self.SetTitle(wx.GetApp().constants.app_display_name)

    def OnCloseWindow(self, event):
        self.parent.Show()
        self.parent.Raise()

        event.Skip()
        super().OnCloseWindow(event)

    def SetTextFont(self, font=None):
        # override to get around float/int bug in 4.1.1!
        if font is None:
            self._textfont = wx.Font(1, wx.FONTFAMILY_SWISS, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_BOLD, False)  # noqa
            self._textsize = int(10)  # noqa
            self._textfont.SetPointSize(self._textsize)

        super().SetTextFont(
            self._textfont
        )


class StatusbarPop(wx.Panel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)  # *args, **kwargs
        self.parent = self.GetParent()
        self.should_show = False
        self.has_message = False
        self.window = self.GetTopLevelParent()

        self.SetSize(wx.Size(-1, 20))
        self.Show(False)

        # root sizer
        root_sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(root_sizer)

        # top border
        self.border_top = wx.Panel(
            self,
            size=wx.Size(-1, 1)
        )
        root_sizer.Add(
            self.border_top,
            0,
            wx.EXPAND,
            0
        )

        # sub sizer
        sub_sizer = wx.BoxSizer(wx.HORIZONTAL)
        root_sizer.Add(
            sub_sizer,
            1,
            wx.EXPAND
        )

        # left spacer
        sub_sizer.Add(
            wx.Size(0, 0),
            0,
            wx.LEFT,
            2
        )

        # text field
        self.text_field = wx.StaticText(self)
        sub_sizer.Add(
            self.text_field,
            1,
            wx.ALIGN_CENTER | wx.ALL,
            2
        )
        self.text_field.SetFont(
            wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        )

        # right spacer
        sub_sizer.Add(
            wx.Size(0, 0),
            0,
            wx.RIGHT,
            2
        )

        # right border
        self.border_right = wx.Panel(
            self,
            size=wx.Size(1, -1)
        )
        sub_sizer.Add(
            self.border_right,
            0,
            wx.EXPAND,
            0
        )

        self.GetParent().Bind(
            wx.EVT_SIZE, self.handle_resize
        )

    def InheritAttributes(self):
        return False

    def SetLabel(self, label):
        self.text_field.SetLabel(label)

        if label == "":
            self.has_message = False
        else:
            self.has_message = True

        def trigger_view_change():
            self.Show(
                self.has_message and self.should_show
            )
            self.Layout()
            self.Refresh()

        if self.has_message or not self.should_show:
            trigger_view_change()
        else:
            wx.CallLater(500, trigger_view_change)

    def handle_resize(self, event=None):
        if event is not None:
            event.Skip()

        parent_size = self.window.ToDIP(
            self.GetParent().GetSize()
        )

        self.SetSize(
            self.window.FromDIP(
                wx.Size(
                    round(parent_size[0] / 4),
                    20
                )
            )
        )

        self.SetPosition(
            self.window.FromDIP(
                wx.Point(
                    0,
                    parent_size[1] - 20
                )
            )
        )

    def SetBackgroundColour(self, color):
        super().SetBackgroundColour(color)
        self.text_field.SetBackgroundColour(color)
        self.Update()

    def SetForegroundColour(self, color):
        super().SetForegroundColour(color)
        self.text_field.SetForegroundColour(color)
        self.Update()

    def SetBorderColour(self, color):  # noqa
        self.border_top.SetBackgroundColour(color)
        self.border_right.SetBackgroundColour(color)


class ThemeChangeDetect(wx.Timer):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent

        # fire on a 200ms loop
        self.Start(200)

    def Notify(self):
        is_dark = self.parent.check_is_dark()
        gen_icons = self.parent.config.ReadBool("view/gen_icons")
        theme = self.parent.config.ReadInt("view/theme")
        fs_theme = self.parent.config.ReadInt("view/fs_theme")

        if (
            self.parent.is_dark != is_dark or
            self.parent.theme != theme or
            self.parent.fs_theme != fs_theme or
            self.parent.using_gen_icons != gen_icons
        ):
            self.parent.is_dark = is_dark
            self.parent.theme = theme
            self.parent.fs_theme = fs_theme
            self.parent.using_gen_icons = gen_icons
            super().Notify()


class WindowThemeable(wx.Frame):
    def __init__(self, *args, **kwargs):  # noqa
        # every window should have the app and translator
        self.app = wx.GetApp()
        self._ = self.app.constants.translation.gettext

        self.handle = self.GetHandle()
        self.theme_colors = {}
        self.gtk_css = None

        color_codes = [
            "workspace",
            "window",
            "window_text",
            "menu_bar",
            "menu_text",
            "menu_highlight",
            "widget",
            "disabled",
            "disabled_text",
            "toolbars",
            "separator",
            "highlight",
            "link_text"
        ]
        for this_color in color_codes:
            self.theme_colors[this_color] = wx.Colour()

        if wx.Platform == "__WXMSW__":
            # tell Windows that dark mode is allowed in this window
            ctypes.windll.uxtheme[133](  # AllowDarkModeForWindow (uxtheme@133)
                self.handle,
                True
            )

        self.Bind(wx.EVT_SHOW, self.do_theme_change)

    # provide FromDIP/ToDIP methods if we don't already have them (i.e., support older wxPython)
    def FromDIP(self, size):
        if (
            hasattr(super(), "FromDIP") and
            callable(super().FromDIP)
        ):
            return super().FromDIP(size)
        else:
            return size

    def ToDIP(self, size):
        if (
            hasattr(super(), "ToDIP") and
            callable(super().ToDIP)
        ):
            return super().ToDIP(size)
        else:
            return size

    def set_color_array(self, override=None, invert=False):
        if override is not None:
            is_dark = override
        else:
            is_dark = self.app.is_dark

        # has the user set the setting to use generic theme colors?
        generic_colors = self.app.config.ReadBool("view/gen_colors")

        # In two cases we need to force generic colors even if the user hasn't checked that setting.
        # First, if this is Windows AND we need dark mode...because win32api doesn't have a native dark mode.
        # Second, if we're in GTK AND we're forcing inverted from the system dark or light mode.
        if (
            (self.app.platform == "win" and is_dark) or
            (wx.Platform == "__WXGTK__" and invert)
        ):
            generic_colors = True

        # workspace color is generic WX gray (only seen as BG for mainwin in full screen mode)
        self.theme_colors["workspace"].Set(
            "GREY"
        )

        # we need to define the theme colors
        if generic_colors:
            # we're using generic colors instead of the usual system colors
            # Based on Bootstrap 5.3 default theme: https://getbootstrap.com/docs/5.3/customize/color/

            if is_dark:
                self.theme_colors["window"].Set("#2c3035")  # bs-tertiary-bg
                self.theme_colors["window_text"].Set("#aeb5bc")  # bs-body-color
                self.theme_colors["menu_bar"].Set("#353a40")  # bs-secondary-bg
                self.theme_colors["menu_text"].Set("#aeb5bc")  # bs-body-color
                self.theme_colors["menu_highlight"].Set("#081532")  # bs-primary-bg-subtle
                self.theme_colors["widget"].Set("#222529")  # bs-body-bg
                self.theme_colors["disabled"].Set("#2c3035")  # bs-tertiary-bg
                self.theme_colors["disabled_text"].Set("#909293")  # bs-tertiary-color
                self.theme_colors["toolbars"].Set("#353a40")  # bs-secondary-bg
                self.theme_colors["separator"].Set("#4a5056")  # bs-border-color
                self.theme_colors["highlight"].Set("#081532")  # bs-primary-bg-subtle
                self.theme_colors["link_text"].Set("#3668f7")  # bs-primary
            else:
                self.theme_colors["window"].Set("#f8f9fa")  # bs-tertiary-bg
                self.theme_colors["window_text"].Set("#222529")  # bs-body-color
                self.theme_colors["menu_bar"].Set("#e9ecef")  # bs-secondary-bg
                self.theme_colors["menu_text"].Set("#222529")  # bs-body-color
                self.theme_colors["menu_highlight"].Set("#d3e1fd")  # bs-primary-bg-subtle
                self.theme_colors["widget"].Set("#ffffff")  # bs-body-bg
                self.theme_colors["disabled"].Set("#f8f9fa")  # bs-tertiary-bg
                self.theme_colors["disabled_text"].Set("#686d73")  # bs-tertiary-color
                self.theme_colors["toolbars"].Set("#e9ecef")  # bs-secondary-bg
                self.theme_colors["separator"].Set("#ced4da")  # bs-dark-bg-subtle
                self.theme_colors["highlight"].Set("#d3e1fd")  # bs-primary-bg-subtle
                self.theme_colors["link_text"].Set("#3668f7")  # bs-primary

        elif self.app.platform == "hku":
            # Haiku does not (yet?) include a GTK theme that matches native colors and widgets. There are a few
            # available online, but they're either not very good or they're copyleft licensed, which would effectively
            # re-license us under copyleft, which we do not want. For now, we're manually setting colors based on the
            # Haiku default values: https://github.com/haiku/haiku/blob/master/src/kits/interface/InterfaceDefs.cpp

            if is_dark:
                self.theme_colors["window"].Set(43, 43, 43)  # Panel background
                self.theme_colors["window_text"].Set(253, 253, 253)  # Panel text
                self.theme_colors["menu_bar"].Set(28, 28, 28)  # Menu background
                self.theme_colors["menu_text"].Set(255, 255, 255)  # Menu item text
                self.theme_colors["menu_highlight"].Set(90, 90, 90)  # Selected menu item background
                self.theme_colors["widget"].Set(29, 29, 29)  # Control background
                self.theme_colors["disabled"].Set(50, 50, 50)  # Inactive window tab
                self.theme_colors["disabled_text"].Set(255, 255, 255)  # Inactive window tab text
                self.theme_colors["toolbars"].Set(28, 28, 28)  # Menu background
                self.theme_colors["separator"].Set(195, 195, 195)  # Control border
                self.theme_colors["highlight"].Set(90, 90, 90)  # Selected menu item background
                self.theme_colors["link_text"].Set(106, 112, 212)  # Link text
            else:
                self.theme_colors["window"].Set(216, 216, 216)  # Panel background
                self.theme_colors["window_text"].Set(0, 0, 0)  # Panel text
                self.theme_colors["menu_bar"].Set(216, 216, 216)  # Menu background
                self.theme_colors["menu_text"].Set(0, 0, 0)  # Menu item text
                self.theme_colors["menu_highlight"].Set(153, 153, 153)  # Selected menu item background
                self.theme_colors["widget"].Set(245, 245, 245)  # Control background
                self.theme_colors["disabled"].Set(232, 232, 232)  # Inactive window tab
                self.theme_colors["disabled_text"].Set(80, 80, 80)  # Inactive window tab text
                self.theme_colors["toolbars"].Set(216, 216, 216)  # Menu background
                self.theme_colors["separator"].Set(172, 172, 172)  # Control border
                self.theme_colors["highlight"].Set(153, 153, 153)  # Selected menu item background
                self.theme_colors["link_text"].Set(51, 102, 187)  # Link text

        else:
            # in other cases, we'll use the system colors (with a few little tweaks)
            self.theme_colors["window"].SetRGBA(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_FRAMEBK).GetRGBA()
            )
            self.theme_colors["window_text"].SetRGBA(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOWTEXT).GetRGBA()
            )
            self.theme_colors["menu_highlight"].SetRGBA(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_MENUBAR).GetRGBA()
            )
            self.theme_colors["widget"].SetRGBA(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW).GetRGBA()
            )
            self.theme_colors["disabled_text"].SetRGBA(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOWTEXT)
                .ChangeLightness(105)
                .GetRGBA()
            )
            self.theme_colors["highlight"].SetRGBA(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNHIGHLIGHT).GetRGBA()
            )
            self.theme_colors["link_text"].SetRGBA(
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_HOTLIGHT).GetRGBA()
            )

            # now we have some oddballs to deal with
            if wx.Platform == "__WXMAC__":
                # first, we have some that need to be different on Mac
                if self.app.is_dark:
                    dis_bg = 105
                else:
                    dis_bg = 95

                self.theme_colors["toolbars"].SetRGBA(
                    wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE).GetRGBA()
                )
                self.theme_colors["disabled"].SetRGBA(
                    wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW)
                    .ChangeLightness(dis_bg)
                    .GetRGBA()
                )
                self.theme_colors["separator"].SetRGBA(
                    wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOWFRAME).GetRGBA()
                )
            else:
                # but they're the same on all the other platforms
                self.theme_colors["toolbars"].SetRGBA(
                    wx.SystemSettings.GetColour(wx.SYS_COLOUR_MENUBAR).GetRGBA()
                )
                self.theme_colors["disabled"].SetRGBA(
                    wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNFACE).GetRGBA()
                )
                self.theme_colors["separator"].SetRGBA(
                    wx.SystemSettings.GetColour(wx.SYS_COLOUR_BTNSHADOW).GetRGBA()
                )

            # now check for a color mismatch situation (light theme with dark menus)
            if (
                    not self.app.is_dark and
                    self.app.check_is_dark(
                        self.theme_colors["menu_bar"],
                        self.theme_colors["menu_text"]
                    )
            ):
                # in mismatch case, our menu color values (for toolbar/statusbar) should be window values
                self.theme_colors["menu_bar"].SetRGBA(
                    self.theme_colors["window"].GetRGBA()
                )
                self.theme_colors["menu_text"].SetRGBA(
                    self.theme_colors["window_text"].GetRGBA()
                )
            else:
                # otherwise, the normal settings
                self.theme_colors["menu_bar"].SetRGBA(
                    wx.SystemSettings.GetColour(wx.SYS_COLOUR_MENUBAR).GetRGBA()
                )
                self.theme_colors["menu_text"].SetRGBA(
                    wx.SystemSettings.GetColour(wx.SYS_COLOUR_MENUTEXT).GetRGBA()
                )

        return self.theme_colors

    def do_theme_change(self, event=None, show=False, workspace=None, recurse=True, full_screen=None):
        if event is not None and not event.IsShown():
            return False

        # correct the "show" value
        show = show or (event is not None and event.IsShown())

        theme_list = self.app.constants.themes
        theme_setting_code = self.app.config.ReadInt("view/theme")
        theme_setting_code_fs = self.app.config.ReadInt("view/fs_theme")
        app_dark = self.app.is_dark
        color_override = None
        need_invert = False
        force_gen_icons = False

        # if the fullscreen setting is just to follow the main one, reset full_screen to false
        if theme_setting_code_fs < 0:
            full_screen = False

        if full_screen is None:
            full_screen = self.IsFullScreen()

        # first, what mode is being requested here? (based on settings and fullscreen status)
        # initial assumption will be the main theme setting
        requested_theme = theme_list[theme_setting_code][0]

        # but wait, what if we're fullscreen? see if we have a setting [i.e., not (none)]
        if full_screen and theme_setting_code_fs > -1:
            requested_theme = theme_list[theme_setting_code_fs][0]

        # if the requested theme isn't "auto" we will need to force it
        force = (requested_theme != "auto")
        to_dark = (requested_theme == "dark")

        # Mac and GTK (Linux) are designed to have manual toggles, so we need to throw them first (or clear them if
        # we're in "auto" mode) so dark mode detection will work correctly later. This isn't needed on Windows because
        # there's really no such thing as dark mode in win32api apps. It's all fake and isn't app-aware.
        if self.app.platform == "mac":
            # on Macs, automatic mode is...automatic. But we need to override if the user picked a preferred mode.
            mac_app = Cocoa.NSApp
            if force:
                if to_dark:
                    mac_appearance = Cocoa.NSAppearance.appearanceNamed_("NSAppearanceNameDarkAqua")
                else:
                    mac_appearance = Cocoa.NSAppearance.appearanceNamed_("NSAppearanceNameAqua")
            else:
                mac_appearance = None

            mac_app.setAppearance_(mac_appearance)

        elif wx.Platform == "__WXGTK__":
            gtk_sett = Gtk.Settings.get_default()

            # baseline is to reset all the GTK stuff to defaults
            gtk_sett.reset_property("gtk-application-prefer-dark-theme")
            gtk_sett.reset_property("gtk-theme-name")
            gtk_sett.reset_property("gtk-icon-theme-name")

            if self.app.platform == "hku":
                if self.gtk_css is None:
                    self.gtk_css = {
                        "provider": Gtk.CssProvider(),
                        "context": Gtk.StyleContext(),
                        "screen": Gdk.Screen.get_default()
                    }
                    self.gtk_css["context"].add_provider_for_screen(
                        self.gtk_css["screen"],
                        self.gtk_css["provider"],
                        Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
                    )
                self.gtk_css["provider"].load_from_data(b'')

                # we need to get the Haiku system font & size to try to match the UI a bit better
                try:
                    settings_file = pathlib.Path(
                        str(
                            subprocess.run(
                                ["finddir", "B_USER_SETTINGS_DIRECTORY"],
                                capture_output=True
                            ).stdout.decode("UTF-8").strip()
                        ) + "/system/app_server/fonts"
                    )
                    settings_load = str(
                        subprocess.run(
                            ["message", settings_file.resolve()],
                            capture_output=True
                        ).stdout.decode("UTF-8").strip()
                    )

                    font = re.search(
                        r'(plain family = string\(")(.*?)(",)',
                        settings_load
                    ).group(2)

                    size = round(
                        float(
                            re.search(
                                r'(plain size = float\()(.*?)(\))',
                                settings_load
                            ).group(2)
                        ) * 0.75
                    )
                except:  # noqa
                    # fallback/default haiku font
                    font = "Noto Sans Display"
                    size = 9

                gtk_sett.set_property("gtk-font-name", font + " " + str(size))

                # force to Adwaita, which is the base of our Haiku customizations
                gtk_sett.set_property("gtk-theme-name", "Adwaita")

                # permanent scrollbars with no hover nonsense (the way things used to be ;-))
                gtk_sett.set_property("gtk-overlay-scrolling", False)

            if force:
                # remember our base theme (i.e., system default)
                sys_theme = gtk_sett.get_property("gtk-theme-name")
                sys_theme_is_dark = self.app.check_is_really_dark()

                # set whether the app should prefer a dark theme (or not)
                gtk_sett.set_property(
                    "gtk-application-prefer-dark-theme",
                    to_dark
                )

                # now we have to force GTK to refresh color values (set theme to its previous value)
                # here is where we'll check whether the system theme supports our current mode
                gtk_sett.set_property("gtk-theme-name", sys_theme)

                # will we need to pick some new icons? (i.e., base theme is opposite of requested theme)
                need_new_icons = (
                    to_dark != sys_theme_is_dark
                )

                # do we need to invert our colors?
                need_invert = (
                    to_dark != self.app.check_is_really_dark()
                )

                # get our dark flags, we might need them!
                dark_flags = self.app.constants.dark_flags

                if need_invert:
                    sys_theme_list = []
                    sys_theme_candidate = None

                    # theme directory list generation borrowed from the Linux Mint Cinnamon project:
                    # https://github.com/linuxmint/cinnamon/blob/master/files/usr/share/cinnamon/cinnamon-settings/modules/cs_themes.py
                    theme_dirs = [
                        os.path.join(GLib.get_home_dir(), ".themes"),
                        os.path.join(GLib.get_user_data_dir(), "themes")
                    ] + [os.path.join(datadir, "themes") for datadir in GLib.get_system_data_dirs()]

                    for this_dir in theme_dirs:
                        dir_path = pathlib.Path(this_dir)
                        if dir_path.is_dir():
                            for theme_path in dir_path.iterdir():
                                theme_name = theme_path.name
                                gtk_dir = pathlib.Path(str(theme_path) + "/gtk-3.0")
                                if gtk_dir.is_dir() and theme_name not in sys_theme_list:
                                    sys_theme_list.append(theme_name)

                    if to_dark:
                        # normal theme doesn't support dark, but we're going to dark
                        possibles = []
                        for item in dark_flags:
                            possibles.append(
                                sys_theme + item
                            )

                        # we have to insert dark in the Linux Mint pattern
                        rewriter = sys_theme.split("-")
                        rewriter.insert(-1, "Dark")

                        possibles.append("-".join(rewriter))

                        for check in possibles:
                            if check in sys_theme_list:
                                sys_theme_candidate = check
                                break
                    else:
                        # we're going light, but we have a dark icon theme (reasons?)
                        for check in dark_flags:
                            temp_name = sys_theme.replace(check, "")
                            if temp_name != sys_theme and temp_name in sys_theme_list:
                                sys_theme_candidate = temp_name
                                break

                    if sys_theme_candidate is None:
                        fallback_theme = "Adwaita"
                        fallback_dark_theme = "Adwaita-dark"

                        if to_dark and fallback_dark_theme in sys_theme_list:
                            sys_theme_candidate = fallback_dark_theme
                        elif not to_dark and fallback_theme in sys_theme_list:
                            sys_theme_candidate = fallback_theme

                    # set the chosen icon theme
                    if sys_theme_candidate is not None:
                        # set the theme
                        gtk_sett.set_property("gtk-theme-name", sys_theme_candidate)

                        # now we don't need to force invert colors later
                        need_invert = False

                if need_new_icons:
                    sys_icon_theme = gtk_sett.get_property("gtk-icon-theme-name")
                    sys_icon_theme_list = {}
                    sys_icon_candidate = None

                    # theme directory list generation borrowed from the Linux Mint Cinnamon project:
                    # https://github.com/linuxmint/cinnamon/blob/master/files/usr/share/cinnamon/cinnamon-settings/modules/cs_themes.py
                    icon_dirs = [
                        os.path.join(GLib.get_home_dir(), ".icons"),
                        os.path.join(GLib.get_user_data_dir(), "icons")
                    ] + [os.path.join(datadir, "icons") for datadir in GLib.get_system_data_dirs()]

                    for this_dir in icon_dirs:
                        dir_path = pathlib.Path(this_dir)
                        if dir_path.is_dir():
                            for theme_path in dir_path.iterdir():
                                theme_conf_file = pathlib.Path(
                                    str(theme_path.resolve()) + "/index.theme"
                                )
                                if theme_conf_file.is_file():
                                    theme_conf = configparser.ConfigParser(
                                        strict=False
                                    )
                                    theme_conf.read(
                                        str(theme_conf_file.resolve())
                                    )

                                    this_theme_name = theme_conf_file.parent.name
                                    this_theme_color_support = theme_conf.getboolean(
                                        "Icon Theme",
                                        "FollowsColorScheme",
                                        fallback=False
                                    )

                                    if this_theme_name is not None:
                                        sys_icon_theme_list[this_theme_name] = this_theme_color_support

                    if sys_icon_theme_list[sys_icon_theme]:
                        sys_icon_candidate = sys_icon_theme
                    elif to_dark:
                        # normal theme doesn't support dark, but we're going to dark
                        possibles = []
                        for item in dark_flags:
                            possibles.append(
                                sys_icon_theme + item
                            )

                        # we have to insert dark in the Linux Mint pattern
                        rewriter = sys_icon_theme.split("-")
                        rewriter.insert(-1, "Dark")

                        possibles.append("-".join(rewriter))

                        for check in possibles:
                            if check in sys_icon_theme_list:
                                sys_icon_candidate = check
                                break
                    else:
                        # we're going light, but we have a dark icon theme (reasons?)
                        for check in dark_flags:
                            temp_name = sys_icon_theme.replace(check, "")
                            if temp_name != sys_icon_theme and temp_name in sys_icon_theme_list:
                                sys_icon_candidate = temp_name
                                break

                    # set the chosen icon theme
                    if sys_icon_candidate is not None:
                        gtk_sett.set_property("gtk-icon-theme-name", sys_icon_candidate)
                    else:
                        # if we didn't find good replacement icons for this theme, use the generic ones
                        force_gen_icons = True

        # now let's set if we're in dark mode or not and set the colors accordingly
        should_dark = self.app.check_is_dark(fs_override=full_screen)

        if should_dark != app_dark:
            color_override = should_dark

        # do we have a parent window? if so, just grab its colors
        this_parent = self.GetParent()
        if this_parent is not None:
            colors = self.theme_colors = this_parent.theme_colors
        else:
            colors = self.set_color_array(
                override=color_override,
                invert=need_invert
            )

        if workspace is not None:
            self.SetBackgroundColour(colors["workspace"])
            self.SetForegroundColour(colors["window_text"])
            workspace.SetBackgroundColour(colors["workspace"])
            workspace.SetForegroundColour(colors["window_text"])
        else:
            self.SetBackgroundColour(colors["window"])
            self.SetForegroundColour(colors["window_text"])

        if self.app.platform == "win":
            # Windows imitation dark mode - so dumb that this has to be done manually.
            #
            # win32api does not support this except through undocumented methods. That's extremely annoying.
            # Hacky/experimental support is in wxWidgets, and presumably coming eventually to wxPython too.
            # Much of this code is derived from the wxWidgets darkmode.cpp. Whenever wxPython starts supporting
            # this officially, we can rip most of this custom junk out. That will be a glorious day.

            # set the window chrome to light or dark (hook into dwmapi.DwmSetWindowAttribute)
            win_is_dark = ctypes.wintypes.BOOL(should_dark)
            ctypes.windll.dwmapi.DwmSetWindowAttribute(
                self.handle,
                20,  # DWMWA_USE_IMMERSIVE_DARK_MODE = 20
                ctypes.byref(win_is_dark),
                ctypes.wintypes.INT(
                    ctypes.sizeof(win_is_dark)
                )
            )

        elif self.app.platform == "hku" and show:
            css_string = pathlib.Path(
                "./commit/assets/haiwaita/haiku-tweaks.css"
            ).read_text()

            def tweak_color(val, tint):
                if tint < 1:
                    return round(255 - (255 - val) * tint)
                else:
                    return round(val * (2 - tint))

            def replacer(match):
                this_color = wx.SystemSettings.GetColour(
                    wx.SYS_COLOUR_WINDOWTEXT
                )
                this_key = (match.group(2)).split(":")

                if this_key[0] in self.theme_colors and self.theme_colors[this_key[0]].IsOk():
                    # the unmodified color
                    this_color.SetRGBA(
                        self.theme_colors[this_key[0]].GetRGBA()
                    )

                    # a tinted color was requested, so we need to adjust (this can involve multiple jumps)
                    if len(this_key) > 1:
                        for my_tint in this_key[1:]:
                            my_tint = float(my_tint)
                            this_color.Set(
                                tweak_color(this_color.Red(), my_tint),
                                tweak_color(this_color.Green(), my_tint),
                                tweak_color(this_color.Blue(), my_tint)
                            )

                return this_color.GetAsString(wx.C2S_CSS_SYNTAX)

            css_string = re.sub(r"(\[\[)(.*?)(\]\])", replacer, css_string)  # noqa

            self.gtk_css["provider"].load_from_data(
                css_string.encode()
            )

        # Okey-dokey - now iterate through and change things if recurse == True
        if self.app.platform == "win" and recurse:
            supported_classes_basic = [
                "CheckBox",
                "StaticText"
            ]
            supported_classes_input = [
                "TextCtrl",
                "SpinCtrl"
            ]

            def recurse_controls(window):
                is_changed = False

                class_name = window.__class__.__name__
                if class_name in supported_classes_basic:
                    # for basic classes we just set fore/background colors
                    window.SetBackgroundColour(colors["window"])
                    window.SetForegroundColour(colors["window_text"])
                    is_changed = True
                elif class_name in supported_classes_input:
                    window.SetBackgroundColour(colors["widget"])
                    window.SetForegroundColour(colors["window_text"])
                    is_changed = True
                elif class_name == "Panel" and window.GetSize()[1] == 1:
                    # 1px high panels are assumed to be separators
                    window.SetBackgroundColour(colors["separator"])
                    is_changed = True
                elif class_name == "HyperLinkBtn":
                    window.SetBackgroundColour(colors["window"])
                    window.SetForegroundColour(colors["link_text"])
                    # Haiku has a problem with CURSOR_HAND, so we use PENCIL, which...is the hand....
                    if self.app.platform == "hku":
                        window.SetLinkCursor(wx.CURSOR_PENCIL)
                    is_changed = True

                if class_name == "Button":
                    if self.app.is_dark:
                        window.Bind(wx.EVT_ENTER_WINDOW, self.handle_button_mouseover)
                        window.Bind(wx.EVT_LEAVE_WINDOW, self.handle_button_mouseover)
                    else:
                        window.Unbind(wx.EVT_ENTER_WINDOW)
                        window.Unbind(wx.EVT_LEAVE_WINDOW)

                if is_changed:
                    window.Refresh()

                for sub_win in window.GetChildren():
                    recurse_controls(sub_win)

            recurse_controls(self)

        self.Refresh()

        if event is not None:
            event.Skip()

        return (
            should_dark,
            force_gen_icons
        )

    def handle_button_mouseover(self, event):
        event.Skip()
        this_obj = event.GetEventObject()
        switch_evts = [
            wx.wxEVT_ENTER_WINDOW,
            wx.wxEVT_SET_FOCUS
        ]

        if event.GetEventType() in switch_evts:
            this_obj.SetForegroundColour(self.theme_colors["widget"])
        else:
            this_obj.SetForegroundColour(self.theme_colors["window_text"])
