import pathlib          # Python Pathlib - https://docs.python.org/3/library/pathlib.html
import re               # Python Regex - https://docs.python.org/3/library/re.html

import wx               # wxPython - https://wxpython.org/
import wx.lib.busy      # wxPython Busy - https://docs.wxpython.org/wx.lib.busy.BusyInfo.html

import markdown         # Python-Markdown - https://github.com/Python-Markdown/markdown

try:
    import pypandoc     # pypandoc - https://github.com/JessicaTegner/pypandoc # noqa
    pypandoc_present = True
except:  # noqa
    pypandoc_present = False
try:
    pypandoc.get_pandoc_path()
    pandoc_present = True
except:  # noqa
    pandoc_present = False
    from bs4 import BeautifulSoup  # noqa


class FileConvert:
    def __init__(self):
        # get app obj
        self.app = wx.GetApp()
        self._ = self.app.constants.translation.gettext

        # get the Pandoc and PyPandoc statuses
        self.has_pypandoc = pypandoc_present
        self.has_pandoc = pandoc_present

        # default arguments for pandoc conversions can go here
        self.args_import = ["--wrap=none"]
        self.args_export = ["--wrap=auto", "--columns=100"]

    def pandoc_check(self, window, should_bypass=False):
        # Return true if Pandoc is present and accounted for
        if self.has_pandoc:
            return True

        # Return false if bypass requested or no pypandoc
        if (
            should_bypass or
            not self.has_pypandoc or
            not window.app.config.ReadBool("params/pandoc_check", True)
        ):
            return False

        # we have pypandoc, but no pandoc - try installing it!

        # build dialog
        pandoc_dialog = wx.MessageDialog(
            window,
            self._(
                "Advanced import and export features require the installation of Pandoc, an open source document "
                "conversion tool. Commit can attempt to automatically download and install Pandoc. Would you "
                "like to proceed?"
            ),
            self._("Install Pandoc?"),
            wx.YES_NO | wx.YES_DEFAULT | wx.ICON_INFORMATION
        )

        # ask?
        confirm_result = pandoc_dialog.ShowModal()

        # handle the answers
        if confirm_result == wx.ID_YES:
            with wx.lib.busy.BusyInfo(
                "Attempting to install Pandoc. Please wait...",
                window,
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW),
                wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOWTEXT)
            ):
                try:
                    pypandoc.pandoc_download.download_pandoc(
                        delete_installer=True,
                        download_folder=str(self.app.config.path)
                    )
                    self.has_pandoc = True
                    return True
                except OSError:
                    pass

            wx.MessageBox(
                self._(
                    "Unable to install Pandoc automatically. Please install it manually to use advanced "
                    "import and export features."
                ),
                self._("Error"),
                wx.OK | wx.ICON_ERROR,
                window
            )

        window.app.config.WriteBool("params/pandoc_check", False)
        return False

    # get supported formats and return the list
    def get_supported_formats(self, fmt_type="import"):
        # supported format types = import, export_file, export_clip

        formats = self.app.constants.conversion_formats
        fmt_supported = ["txt"]  # txt is ALWAYS supported, pandoc or not, import or export
        format_names = []
        format_exts = []
        ext_to_pandoc = {}
        dialog_string = ""
        all_formats_string = ""

        if self.has_pandoc:
            pandoc_from_formats, pandoc_to_formats = pypandoc.get_pandoc_formats()
            if fmt_type == "import":
                fmt_supported.extend(pandoc_from_formats)
            else:
                fmt_supported.extend(pandoc_to_formats)
        else:
            # fallback types (works without pandoc)
            fmt_supported.extend(["html", "commonmark_x"])

        for fmt, detail in formats.items():
            if detail[fmt_type] and fmt in fmt_supported:
                format_names.append(detail["name"])
                format_exts.append(detail["extension"])
                formatted_ext = "*" + detail["extension"]

                # add to the reverse mapping (extension to pandoc code)
                ext_to_pandoc[detail["extension"]] = fmt

                # add to the dialog format string
                dialog_string += detail["name"] + " (" + formatted_ext + ")|" + formatted_ext + "|"

                # add to the all formats string
                if all_formats_string != "":
                    all_formats_string += ";"
                all_formats_string += formatted_ext

        if fmt_type == "import":
            dialog_string = (
                "Supported formats...|" + all_formats_string + "|" +
                dialog_string +
                "All files... (*.*)|*.*"
            )
        else:
            dialog_string = dialog_string[0:-1]

        # send back to caller
        return {
            "names": format_names,
            "exts": format_exts,
            "ext_to_pandoc": ext_to_pandoc,
            "dialog_string": dialog_string
        }

    # do an import
    def do_import_file(self, window):
        use_pandoc = self.pandoc_check(window)
        formats = self.get_supported_formats("import")

        # build the import dialog
        import_dialog = wx.FileDialog(
            window,
            self._("Import from file..."),
            wildcard=formats["dialog_string"],
            defaultDir=str(window.file.find_default_dir()),
            style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST
        )
        import_dialog.SetFilterIndex(0)
        if import_dialog.ShowModal() == wx.ID_CANCEL:
            return
        
        # get file info via pathlib
        file = pathlib.Path(import_dialog.GetPath())
        ext = file.suffix

        # unrecognized extension? allow user to manually select a type
        if ext not in formats["exts"]:
            format_dialog = wx.SingleChoiceDialog(
                window,
                self._("Please select a supported file format for this import:"),
                self._("Unknown file type..."),
                formats["names"]
            )
            format_dialog.SetSize(
                wx.Size(300, 300)
            )
            format_dialog.Center()
            if format_dialog.ShowModal() == wx.ID_CANCEL:
                return

            ext = formats["exts"][format_dialog.GetSelection()]

        if use_pandoc:
            # use pandoc to open up supported document types (if we have pandoc)
            try:
                # txt needs special handling
                if ext == ".txt":
                    output = pypandoc.convert_text(
                        self.txt_to_md(file.read_text()),
                        format="commonmark_x",
                        to="commonmark_x-raw_html",
                        outputfile=None,  # returning as a string in this case
                        extra_args=self.args_import
                    )
                else:
                    output = pypandoc.convert_file(
                        str(file.absolute()),
                        format=formats["ext_to_pandoc"][ext],
                        to="commonmark_x-raw_html",
                        outputfile=None,
                        extra_args=self.args_import
                    )
            except Exception:  # NOQA
                wx.MessageBox(
                    self._("Unable to import the specified file."),
                    self._("Error"),
                    wx.OK | wx.ICON_ERROR
                )
                return False

        else:
            # no pandoc, but we can still do some things another way
            if ext == ".md":
                # Just pull Markdown in directly; will normalize breaks below
                output = self.md_to_ctxt(
                    file.read_text()
                )
            elif ext == ".txt":
                output = self.txt_to_md(
                    file.read_text()
                )
            elif ext == ".html":
                html_parse = BeautifulSoup(
                    file.read_text(),
                    "lxml"
                )
                output = self.txt_to_md(
                    html_parse.get_text()
                )
            else:
                return False

        # normalize the line breaks!
        output = self.normalize_line_breaks(output)

        # send the imported content to the editor
        window.file.open(None, is_import=True, raw_text=output)

    # do the export conversion; this is used by all the export functions
    def do_export(self, window, target_format="html", out_file=None, meta=True):
        # use pandoc to convert
        if self.has_pandoc:
            try:
                these_args = self.args_export

                # standalone files get some extra stuff
                if out_file is not None:
                    # set to standalone (put at start of args)
                    these_args.insert(0, "-s")

                    # get the title
                    this_title = self._("Commit Document")
                    if window.file.meta["title"] != "":
                        this_title = window.file.meta["title"]

                    # set title value for HTML files
                    if target_format == "html":
                        these_args.append("--metadata=pagetitle:" + this_title)

                    # set title value to Docbook files
                    if target_format == "dbk" or target_format == "epub":
                        these_args.append("--metadata=title:" + this_title)

                # txt needs special handling...first to MD then through the filter
                if target_format == "txt":
                    these_args.append("--wrap=none")
                    output = self.md_to_txt(
                        pypandoc.convert_text(
                            self.ctxt_to_md(
                                window.file.merge_content(
                                    include_metadata=False  # never include metadata in plain text export
                                )
                            ),
                            format="commonmark_x",
                            to="commonmark_x",
                            outputfile=None,  # returning as a string in this case
                            extra_args=these_args
                        )
                    )
                    if out_file is not None:
                        out_file.write_text(output)
                else:
                    # if we're writing to a file, we need the path
                    if out_file is not None:
                        out_path = str(out_file.absolute())
                    else:
                        out_path = None

                    output = pypandoc.convert_text(
                        self.ctxt_to_md(
                            window.file.merge_content(
                                include_metadata=(
                                    meta and
                                    out_file is not None
                                )  # metadata only if file export and not excluded
                            )
                        ),
                        format="commonmark_x",
                        to=target_format,
                        outputfile=out_path,
                        extra_args=these_args
                    )
                if out_file is not None:
                    # we outputted to a file, so no content to return, but return True for success
                    return True

                return output

            except Exception as e:  # NOQA
                import traceback
                traceback.print_exc()
                return False
        else:
            # no pandoc, but we can still do some basic exports
            if target_format == "commonmark_x":
                output = self.ctxt_to_md(
                    window.file.merge_content(
                        include_metadata=(
                            meta and
                            out_file is not None
                        )  # include meta if md FILE export and not excluded
                    )
                )
            elif target_format == "txt":
                output = self.md_to_txt(
                    self.ctxt_to_md(
                        window.file.merge_content(
                            include_metadata=False  # never metadata for TXT export
                        )
                    )
                )
            elif target_format == "html":
                output = markdown.markdown(
                    self.ctxt_to_md(
                        window.file.merge_content(
                            include_metadata=False  # no metadata for this kind of HTML export
                        )
                    )
                )
                if out_file is not None:
                    output = "<html>\n<body>\n" + output + "\n</body>\n</html>\n"
            else:
                output = False

            # save it!
            if out_file is not None and output:
                out_file.write_text(output)
                return True

            return output

    # method for exporting files to supported file formats
    def do_export_file(self, window):
        self.pandoc_check(window)
        formats = self.get_supported_formats("export_file")

        # build the export dialog
        export_dialog = wx.FileDialog(
            window,
            self._("Export to file..."),
            wildcard=formats["dialog_string"],
            defaultDir=str(window.file.find_default_dir()),
            defaultFile=str(window.file.path.stem),
            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT
        )
        if export_dialog.ShowModal() == wx.ID_CANCEL:
            return

        # get file info via pathlib
        file = pathlib.Path(export_dialog.GetPath())
        target_format = formats["ext_to_pandoc"][
            formats["exts"][
                export_dialog.GetFilterIndex()
            ]
        ]

        result = self.do_export(
            window,
            target_format=target_format,
            out_file=file,
            meta=True
        )

        if not result:
            wx.MessageBox(
                self._("Unable to export to the specified format."),
                self._("Error"),
                wx.OK | wx.ICON_ERROR
            )
            return False

        wx.MessageBox(
            self._("File exported successfully."),
            self._("Success"),
            wx.OK | wx.ICON_INFORMATION
        )
        return True

    # method for exporting files in some formats to the clipboard
    def do_export_clipboard(self, window):
        formats = self.get_supported_formats("export_clip")

        format_dialog = wx.SingleChoiceDialog(
            window,
            self._("Select format for export to clipboard:"),
            self._("Export to clipboard..."),
            formats["names"]
        )
        format_dialog.SetSize(
            wx.Size(300, 300)
        )
        format_dialog.Center()
        if format_dialog.ShowModal() == wx.ID_CANCEL:
            return

        target_format = formats["ext_to_pandoc"][
            formats["exts"][
                format_dialog.GetSelection()
            ]
        ]

        result = self.do_export(
            window,
            target_format=target_format,
            out_file=None,
            meta=False
        )

        if not result:
            wx.MessageBox(
                self._("Unable to export to clipboard."),
                self._("Error"),
                wx.OK | wx.ICON_ERROR
            )
            return False

        # clippy, no!
        the_clipboard = wx.Clipboard()
        this_data = wx.DataObjectComposite()

        # save plain text obj
        this_data.Add(
            wx.TextDataObject(result)
        )

        if target_format in ["html"]:
            this_data.Add(
                wx.HTMLDataObject(result),
                True
            )

        if the_clipboard.Open():
            the_clipboard.SetData(this_data)
            the_clipboard.Flush()
            the_clipboard.Close()
            return True

        wx.MessageBox(
            self._("Unable to export to clipboard."),
            self._("Error"),
            wx.OK | wx.ICON_ERROR
        )
        return False

    # method for exporting files in some formats to the clipboard
    def do_export_print(self, window):
        result = self.do_export(
            window,
            target_format="html",
            out_file=None,
            meta=False
        )
        if result:
            return result

        wx.MessageBox(
            self._("Unable to generate print data."),
            self._("Error"),
            wx.OK | wx.ICON_ERROR
        )
        return False

    @staticmethod
    def ctxt_to_md(content):
        # Our native format is not QUITE valid Markdown, which can cause problems.
        # Specifically, we treat single line breaks as line breaks, while Markdown treats them
        # as spaces. We need to add two spaces to EOL to make it work right on export.
        return re.sub(r"(?<!\n)(\n)(?!\n)", r"  \n", content)

    @staticmethod
    def md_to_ctxt(content):
        split_match = re.match(r"^(---\n[\s\S]*?---\n)?([\s\S]*)$", content)

        # save the metadata
        if split_match.group(1) is not None:
            md_meta = split_match.group(1)
        else:
            md_meta = ""

        md_content = split_match.group(2)

        # normalize line breaks to \n
        md_content = re.sub(r"(\r?\n)", "\n", md_content)

        # get rid of leading breaks
        md_content = re.sub(r"^(\n)*", "", md_content)

        # turn all sequences of >=2 breaks to 2
        md_content = re.sub(r"(\n){2,}", "\n\n", md_content)

        # remove all single line breaks not preceded by two spaces
        md_content = re.sub(r"(?<!\n)(?<! {2})(\n)(?!\n)", " ", md_content)

        return md_meta + md_content

    @staticmethod
    def txt_to_md(txt_content):
        # get rid of leading breaks
        txt_content = re.sub(r"^(\n)*", "", txt_content)

        # normalize all line breaks to doubles
        txt_content = re.sub(r"^(\t+)", "", txt_content)
        txt_content = re.sub(r"(\r?\n\t?)+", r"\n\n", txt_content)

        return txt_content

    @staticmethod
    def md_to_txt(content):
        content = "\t" + content  # tab in first paragraph
        content = re.sub(r"(\r?\n)+", r"\n\t", content)  # normalize all line breaks to singles with indent
        return content

    @staticmethod
    def normalize_line_breaks(content):
        content = re.sub(r"(\r?\n){2,}", "\n\n", content)
        content = re.sub(r"(\r?\n)+$", "", content)
        content = re.sub(r"(\\)", "", content)

        return content
