import locale                   # Python Locale - https://docs.python.org/3/library/locale.html
import os                       # Python Misc. OS Interfaces - https://docs.python.org/3/library/os.html
import pathlib                  # Python Pathlib - https://docs.python.org/3/library/pathlib.html
import re                       # Python Regex - https://docs.python.org/3/library/re.html
import webbrowser               # Python WebBrowser - https://docs.python.org/3/library/webbrowser.html

import wx                       # wxPython - https://wxpython.org/
from wx.adv import Sound        # wxPython Sound - https://docs.wxpython.org/wx.adv.Sound.html

from commit.includes.filehandler import FileHandler         # local - file open, save, and path tools
from commit.includes.fileconvert import FileConvert         # local - file conversion (import/export)
from commit.includes.searchfind import SearchFindDialog     # local - searching and finding
from commit.ui.commit_ui_design import MainWin_Base         # local - main window design from wxFormBuilder
from commit.ui.metawin import MetaWin                       # local - metadata window
from commit.ui.aboutwin import AboutWin                     # local - about window
from commit.ui.prefswin import PrefsWin                     # local - preferences window
from commit.includes.customobjects import (                 # local - custom classes
    StatusbarPop,
    WindowThemeable
)

if wx.Platform == "__WXGTK__":
    import gi  # noqa
    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk  # noqa


class MainWin(MainWin_Base, WindowThemeable):
    def __init__(self, parent, start_file, multi_win=False):
        self.menu_bindings = {}

        # call parent inits
        MainWin_Base.__init__(self, parent)
        WindowThemeable.__init__(self, parent)

        # get app obj locally
        app = self.app

        # init must-have values to their defaults
        self.word_count_start = 0
        self.scroll_bypass = False
        self.is_closing = False
        self.init_completed = False
        self.ignore_edits = False
        self.is_dark_mode = False
        self.popup_open = False
        self.is_full_screen = False
        self.display_font = wx.Font()
        self.gtk_text_committed_widget = None
        self.gtk_text_working_widget = None
        self.meta_open = False

        # remove the update option from the menu if we're a no-update build
        if self.app.updater.disabled:
            self.menu_help.Remove(
                self.menu_help_updates.GetId()
            )

        # init translations
        _ = self._

        # init sounds
        self.sound_commit = pathlib.Path("commit/assets/sound/commit.wav").absolute()
        self.sound_pull = pathlib.Path("commit/assets/sound/pull.wav").absolute()

        # file handle object
        self.file = FileHandler(self)

        # import/export object
        self.convert = FileConvert()

        # hide the option to get Pandoc if we already have it
        if (
            not self.convert.has_pypandoc or
            self.convert.pandoc_check(self, should_bypass=True)
        ):
            self.menu_help.Remove(
                self.menu_help_pandoc.GetId()
            )

        # set text field margins - this does not work in GTK; see GTK-specific workaround in event_window_show()
        field_margins = wx.Point(10, -1)
        self.text_committed_widget.SetMargins(field_margins)
        self.text_working_widget.SetMargins(field_margins)

        # insert the fullscreen spacers, defaulting to hidden
        main_panel_sizer = self.GetSizer()
        self.fs_spacer_1 = main_panel_sizer.PrependStretchSpacer(0)
        self.fs_spacer_2 = main_panel_sizer.AddStretchSpacer(0)

        # set window icon
        self.SetIcons(app.constants.icon)

        # set up the transient status bor (for when the main one is hidden)
        self.statusbar_pop = StatusbarPop(self.main_panel)

        # should fullscreen?
        # need to figure it out here, but we actually do it later (has to be after window show)
        should_fullscreen = app.config.ReadBool("window/is_fullscreen")

        # should maximize?
        if app.config.ReadBool("window/is_maximized"):
            self.Maximize(True)

        # get the screen size
        window_size = wx.DisplaySize()

        # determine window size (previous, unless that's weird)
        new_size = [
            app.config.ReadInt("window/size_x"),
            app.config.ReadInt("window/size_y")
        ]
        if new_size[0] > window_size[0]:
            new_size[0] = round(window_size[0] * 0.9)
        if new_size[1] > window_size[1]:
            new_size[1] = round(window_size[1] * 0.9)

        # determine window position (previous, unless that's weird)
        new_pos = [
            app.config.ReadInt("window/pos_x"),
            app.config.ReadInt("window/pos_y")
        ]

        # add an offset if it's a multiwindow
        if multi_win:
            new_pos[0] += 30
            new_pos[1] += 30

        # make any corrections
        if new_pos[0] > window_size[0] or new_pos[0] < 0:
            new_pos[0] = -1
        if new_pos[1] > window_size[1] or new_pos[1] < 0:
            new_pos[1] = -1

        # dooo it!
        self.SetSize(
            new_pos[0],
            new_pos[1],
            new_size[0],
            new_size[1]
        )

        statusbar_font = self.statusbar_1.GetFont()
        statusbar_font.SetPointSize(8)

        if self.app.platform == "mac":
            # on Mac platform

            # remove minimize and maximize (in the Mac standard window menu)
            self.menu_view.Remove(self.menu_view_minimize)
            self.menu_view.Remove(self.menu_view_maximize)

            # change fullscreen shortcut to Mac norm
            mac_fullscreen_accel = wx.AcceleratorEntry()
            mac_fullscreen_accel.FromString("RAWCTRL+CTRL+F")
            self.menu_view_fullscreen.SetAccel(
                mac_fullscreen_accel
            )

            # adjust the statusbar font (8pt is fine elsewhere, but tiny on Mac)
            statusbar_font.SetPointSize(10)

        else:
            # on non-Mac platforms

            # remove the Mac stub "Window" menu
            # we had to have this to make the Mac built-in Window menu position correctly
            self.menu_main.Remove(3)

        # set the statusbar font size
        self.statusbar_1.SetFont(statusbar_font)
        self.statusbar_2.SetFont(statusbar_font)
        self.statusbar_3.SetFont(statusbar_font)

        # remove the DEBUG menu unless we're launching in debug mode
        if not self.app.constants.debug:
            self.menu_main.Remove(
                self.menu_main.FindMenu(_("DEBUG"))
            )

        # build our alternate main menu for Windows (so annoying!)
        self.menu_main_alt.create()

        # and attach our alternate scroll bars (also annoying)
        self.text_committed_scroll_alt.attach_to(
            self.text_committed_widget
        )
        self.text_working_scroll_alt.attach_to(
            self.text_working_widget
        )

        # center if the new position is default values
        if new_pos[0] == -1 or new_pos[1] == -1:
            self.Center()

        # create all the toolbar and statusbar icon objects we'll need
        self.icons = {
            "tb": {
                "chevron_up": wx.Bitmap(),
                "chevron_down": wx.Bitmap(),
                "prefs": wx.Bitmap(),
                "meta": wx.Bitmap(),
                "fullscreen": wx.Bitmap(),
                "fullscreen_back": wx.Bitmap(),
                "menu": wx.Bitmap(),
                "quit": wx.Bitmap()
            },
            "sb": {
                "meta": wx.Bitmap(),
                "theme": wx.Bitmap(),
                "ratio": wx.Bitmap(),
                "menu": wx.Bitmap()
            },
            "sb_dot": {
                "saved": wx.Bitmap(),
                "unsaved": wx.Bitmap()
            }
        }

        # init the theme list menu
        self.menu_theme = wx.Menu()
        for this_theme in self.app.constants.themes:
            self.menu_theme.AppendRadioItem(
                wx.ID_ANY,
                this_theme[1]
            )
        self.menu_theme.FindItemByPosition(
            self.app.config.ReadInt("view/theme")
        ).Check(True)
        self.menu_theme.Bind(wx.EVT_MENU, self.event_statusbar_theme_select)

        # init the fullscreen ratio list menu
        self.menu_ratio = wx.Menu()
        for this_ratio in self.app.constants.fs_ratio_map:
            self.menu_ratio.AppendRadioItem(
                wx.ID_ANY,
                this_ratio[1]
            )
        self.menu_ratio.FindItemByPosition(
            self.app.config.ReadInt("view/fs_ratio")
        ).Check(True)
        self.menu_ratio.Bind(wx.EVT_MENU, self.event_statusbar_ratio_select)

        # by default, don't show the exit or fullscreen tools, or the statusbar icons
        self.tool_unsaved.Show(False)
        self.tool_meta.Show(False)
        self.tool_menu.Show(False)
        self.tool_exit.Show(False)
        self.sb_fs_buttons.Show(False)
        self.sb_unsaved.Show(False)

        # bind scroll events (so we can prevent them if scrolling committed window is disabled)
        self.text_committed_widget.Bind(wx.EVT_SCROLLWIN, self.event_committed_scroll)
        self.text_committed_widget.Bind(wx.EVT_MOUSEWHEEL, self.event_committed_scroll)

        # bind the statusbar events
        self.Bind(wx.EVT_MENU_HIGHLIGHT, self.event_statusbar_update)
        self.Bind(wx.EVT_MENU_CLOSE, self.event_statusbar_update)
        self.Bind(wx.EVT_TOOL_ENTER, self.event_statusbar_update)

        # bind the paste event on the working text field
        self.text_working_widget.Bind(wx.EVT_TEXT_PASTE, self.event_working_paste)

        # bind all keypresses in the window (so we can intercept shortcuts when needed)
        self.Bind(wx.EVT_CHAR_HOOK, self.event_window_keys)

        # set prefs!
        self.do_prefs_update()

        # enable Apple macOS fullscreen capability (if the method exists, since it's newish)
        if (
            hasattr(self, "EnableFullScreenView") and
            callable(self.EnableFullScreenView) and
            hasattr(wx, "EVT_FULLSCREEN")
        ):
            self.EnableFullScreenView(True)
            self.Bind(wx.EVT_FULLSCREEN, self.do_toggle_fullscreen)

        # show the welcome message for 5s at open of first window
        if not multi_win:
            status_message = _("Welcome to") + " " + self.app.constants.app_display_name + " " +\
                             self.app.constants.app_version
            self.do_update_statusbar(status_message)
            wx.CallLater(
                5000,
                self.do_update_statusbar,
                status_text="",
                must_match=status_message
            )

        # focus to the editing window
        self.text_working_widget.SetFocus()

        # should fullscreen? has to be after show
        # we determine it earlier though, so it doesn't get munged by window sizing
        if should_fullscreen:
            self.do_toggle_fullscreen(force=True)

        # save the first of the last opened files to a var
        last_file = pathlib.Path(
            self.app.recents.open_list[0]
        )

        if multi_win:
            # this is a second (or third, or whatever) window, don't open anything at open
            pass
        elif start_file.is_file() and start_file.suffix == ".ctxt":
            # if a file was passed in on the command line; save last to recents and open req file
            app.recents.file_list.insert(0, str(last_file.resolve()))
            self.file.open(start_file.resolve())
        elif app.config.ReadBool("startup/open_last_edited"):
            file = pathlib.Path(app.config.Read("files/current"))
            if file.is_file():
                self.file.open(file.resolve())
        elif app.config.Read("files/current") != "":
            # if we're not opening on startup, last open file to recent files list and remove from current
            app.recents.file_list.insert(0, str(last_file.resolve()))
            app.config.Write("files/current", "")

        # update word count and recents
        self.update_word_count(True)
        self.app.recents.update()

    # Bind Override?
    def Bind(self, *args, **kwargs):
        super().Bind(*args, **kwargs)
        if args[0].typeId == wx.EVT_MENU.typeId and "id" in kwargs:
            self.menu_bindings[kwargs["id"]] = args[1]

    # do a commit
    def do_commit(self, consider_position=False):
        # get the info we need
        selection = self.text_working_widget.GetSelection()
        end_point = self.text_working_widget.GetInsertionPoint()
        content_length = len(self.text_working_widget.GetValue())

        # we don't *really* want to commit if there's no content
        if content_length == 0:
            return

        if self.app.config.ReadBool("options/commit-sounds"):
            self.do_play_sound(self.sound_commit)

        if selection[0] != selection[1]:
            # if we have a selection, that's what we're committing
            # doesn't matter if we pressed "enter" or used a button/menu
            range_to_commit = [selection[0], selection[1]]
        elif end_point > 0 and consider_position:
            # if considering position (i.e., comes from an enter press and not at pos 0)
            # here we commit from 0 to the insertion point
            range_to_commit = [0, end_point]
        else:
            # if we're in position 0, or if we came from a button/menu, but with no selection
            # in this case, we commit the whole working field
            range_to_commit = [0, self.text_working_widget.GetLastPosition()]

        # append the text we're committing
        if self.text_committed_widget.GetValue() != "":
            self.text_committed_widget.AppendText("\n\n")
        self.text_committed_widget.AppendText(
            (self.text_working_widget.GetRange(
                range_to_commit[0],
                range_to_commit[1]
            )).strip()
        )

        # remove the committed text from the working field
        self.text_working_widget.Remove(range_to_commit[0], range_to_commit[1])

        # reset the working field's undo queue
        if (
            hasattr(self.text_working_widget, "EmptyUndoBuffer") and
            callable(self.text_working_widget.EmptyUndoBuffer)
        ):
            self.text_working_widget.EmptyUndoBuffer()

        # scroll to bottom
        self.do_jump_to_bottom()

        # set that we need an update
        self.file.needs_save = True

        # update the committed word count
        self.update_word_count()

        # save the file (if we have autosave enabled AND a filename)
        if self.file.path.is_file() and self.app.config.ReadBool("options/auto-save-commit"):
            self.file.save()
        else:
            self.file.needs_save = True
            self.update_title()

    # do a pull
    def do_pull(self, on_open=False):
        committed_content = self.text_committed_widget.GetValue()
        working_content = self.text_working_widget.GetValue()

        # can't pull if nothing is there!
        if committed_content == "":
            return

        # start with nothing
        pull_text = None

        # find the last paragraph
        re_obj = re.compile(r"(\n\n)(.*?)$")

        # gather the needed information
        pull_result = re_obj.search(committed_content)

        if pull_result is not None and pull_result.group(2) is not None:
            pull_text = pull_result.group(2)
        else:
            re_obj = re.compile(r"^(.*)$")
            pull_result = re_obj.search(committed_content)
            if pull_result is not None and pull_result.group(1) is not None:
                pull_text = pull_result.group(1)

        if pull_result is None or pull_text is None:
            return

        if not on_open and self.app.config.ReadBool("options/commit-sounds"):
            self.do_play_sound(self.sound_pull)

        # first, delete the content we're pulling
        self.text_committed_widget.Remove(pull_result.start(), pull_result.end())

        if working_content is not None and working_content != "":
            # if we have working content, we need to prepend
            self.text_working_widget.SetValue(pull_text + " " + working_content)
        else:
            # the working field is empty; just plop the text in it!
            self.text_working_widget.SetValue(pull_text)

        # reset the working field's undo queue
        if (
            hasattr(self.text_working_widget, "EmptyUndoBuffer") and
            callable(self.text_working_widget.EmptyUndoBuffer)
        ):
            self.text_working_widget.EmptyUndoBuffer()

        # scroll to ends
        self.do_jump_to_bottom()

        # update the committed word count
        self.update_word_count()

        # and mark that we need a save, unless this happened on a new open
        if not on_open:
            self.file.needs_save = True
            self.update_title()

    # jump to the bottom of the committed text field
    def do_jump_to_bottom(self):
        committed = self.text_committed_widget

        # enable scroll_bypass (to get past disabled scrolling)
        self.scroll_bypass = True

        # set committed widget insertion point at the end
        committed.SetInsertionPoint(
            committed.GetLastPosition()
        )
        committed.ShowPosition(
            committed.GetLastPosition()
        )

        if wx.Platform == "__WXGTK__" and self.gtk_text_committed_widget is not None:
            def do_gtk_bottom_jump():
                # for some reason, jumping to bottom doesn't work right in GTK....
                # so we need to use the GTK method on the underlying TextView object
                vert_adj = self.gtk_text_committed_widget.get_vadjustment()
                vert_adj.set_value(
                    vert_adj.get_upper()
                )
            # AND we have a race condition when doing some things (like changing theme), so put in a CallAfter
            wx.CallAfter(do_gtk_bottom_jump)
            wx.CallLater(500, do_gtk_bottom_jump)

        # workaround problem where it's not always REALLY at the bottom (pgdown)
        committed.ScrollPages(1)

        # disable the scroll bypass (i.e., return to normal behavior)
        self.scroll_bypass = False

    # jump to the bottom but only if we're near the bottom already or have scrolling disabled
    # for some reason, Macs don't report scroller position...so we always jump to bottom :-/ annoying
    def do_jump_to_bottom_if_necessary(self):
        if self.app.config.ReadBool("options/disable-scrolling"):
            # if scrolling is disabled, just jump to bottom always
            wx.CallAfter(self.do_jump_to_bottom)
        else:
            committed = self.text_committed_widget
            scroll_range = committed.GetScrollRange(wx.VERTICAL)
            break_pos = scroll_range * .90
            current_pos = committed.GetScrollPos(wx.VERTICAL)

            if scroll_range <= 0 or current_pos > break_pos:
                # stick to the bottom if we're scrolled within 10% of the end of the document
                # or if we're on a platform that doesn't reveal scroll position (WTF Mac?)
                wx.CallAfter(self.do_jump_to_bottom)

    # handle enable/disable for all the edit menu items
    def do_edit_menu_refresh(self, all_disabled=False):
        working = self.text_working_widget
        focused = self.FindFocus()
        if isinstance(focused, wx.TextCtrl):
            selection = focused.GetSelection()
        else:
            selection = [0, 0]
        has_selection = (selection[0] != selection[1])

        # always clear the undo queue on the committed field (force disable in right-click menu)
        if (
            hasattr(self.text_working_widget, "EmptyUndoBuffer") and
            callable(self.text_working_widget.EmptyUndoBuffer)
        ):
            self.text_committed_widget.EmptyUndoBuffer()

        # update undo menu item
        self.menu_edit_undo.Enable(
            working.CanUndo() and
            not all_disabled
        )

        # update cut menu item
        self.menu_edit_cut.Enable(
            has_selection and
            working.CanCut() and
            not all_disabled
        )

        # update copy menu item
        # this one is weird; we can copy from either working or committed
        # we look at the focused obj, not just working, and ignore the all_disabled option
        self.menu_edit_copy.Enable(
            has_selection and
            focused.CanCopy()
        )

        # update paste menu item
        self.menu_edit_paste.Enable(
            self.check_for_text_clipboard() and
            working.CanPaste() and
            not all_disabled
        )

        # update delete menu item
        self.menu_edit_delete.Enable(
            has_selection and
            not all_disabled
        )

    # handle a theme change (light/dark)
    def do_theme_change(self, event=None, show=False, workspace=None, recurse=False, full_screen=None):
        if event is not None and not event.IsShown():
            return False

        self.Freeze()

        if full_screen is None:
            full_screen = self.is_full_screen

        # full_update
        (is_dark, force_gen) =\
            super().do_theme_change(
                event,
                show,
                workspace=self.main_panel,
                recurse=False,
                full_screen=full_screen
            )

        # color table
        colors = self.theme_colors

        # color changes that we need to do in all cases

        # by default, we want to show or hide the top menu seperator based on fullscreen status
        # But we NEVER want to see it on Mac because it's ugly.
        self.border_menu.Show(
            not self.IsFullScreen() and not self.app.platform == "mac"
        )

        # set separator colors
        self.border_menu.SetBackgroundColour(colors["separator"])
        self.border_toolbar_top.SetBackgroundColour(colors["separator"])
        self.border_toolbar_bottom.SetBackgroundColour(colors["separator"])
        self.border_statusbar.SetBackgroundColour(colors["separator"])
        self.statusbar_pop.SetBorderColour(colors["separator"])

        # set committed widget color (disabled text)
        self.text_committed_widget.SetBackgroundColour(colors["disabled"])
        self.text_committed_widget.SetForegroundColour(colors["window_text"])

        # set toolbar and statusbar widget colors
        self.toolbar_widget.SetBackgroundColour(colors["toolbars"])
        self.statusbar_widget.SetBackgroundColour(colors["toolbars"])

        tool_list = [
            self.tool_commit,
            self.tool_pull,
            self.tool_unsaved,
            self.tool_prefs,
            self.tool_meta,
            self.tool_fullscreen,
            self.tool_menu,
            self.tool_exit,
            self.sb_unsaved,
            self.sb_meta_button,
            self.sb_theme_button,
            self.sb_ratio_button,
            self.sb_menu_button
        ]
        for this_tool in tool_list:
            this_tool.set_color_scheme(
                colors["toolbars"],
                colors["window_text"],
                colors["highlight"]
            )

        # set statusbar widget and text field colors
        self.sb_panel_1.SetBackgroundColour(colors["toolbars"])
        self.statusbar_1.SetForegroundColour(colors["menu_text"])
        self.sb_panel_2.SetBackgroundColour(colors["toolbars"])
        self.statusbar_2.SetForegroundColour(colors["menu_text"])
        self.sb_panel_3.SetBackgroundColour(colors["toolbars"])
        self.statusbar_3.SetForegroundColour(colors["menu_text"])
        self.sb_panel_4.SetBackgroundColour(colors["toolbars"])
        self.statusbar_pop.SetForegroundColour(colors["menu_text"])
        self.sb_fs_buttons.SetBackgroundColour(colors["toolbars"])
        self.statusbar_pop.SetBackgroundColour(colors["toolbars"])

        # working text widget
        # for some reason, firing SetForeground Colour on the working widget triggers an EVT_TEXT...
        # WTF? Why why why why why? So we have a flag telling the event handler to ignore the edit,
        # just this once, because this isn't really an edit. Erghhhhhh.
        self.ignore_edits = True
        self.text_working_widget.SetBackgroundColour(colors["widget"])
        self.text_working_widget.SetForegroundColour(colors["window_text"])

        # fix the style for the committed widget (reset all text to default style)
        # on GTK you get weird/wrong text background effects otherwise
        if wx.Platform == "__WXGTK__":
            self.text_committed_widget.SetStyle(
                0,
                self.text_committed_widget.GetLastPosition(),
                self.text_committed_widget.GetDefaultStyle()
            )
            self.text_working_widget.SetStyle(
                0,
                self.text_working_widget.GetLastPosition(),
                self.text_working_widget.GetDefaultStyle()
            )

        self.ignore_edits = False

        # Silly Windows Nonsense, only if dark modes are supported (otherwise, stay with defaults)
        if self.app.platform == "win":
            # change color of alternate menubar (and activate if we're going dark)
            self.menu_main_alt.set_color_scheme(
                colors["menu_bar"],
                colors["menu_text"],
                colors["highlight"]
            )
            self.menu_main_alt.activate(
                is_dark
            )

            # show/hide the real and fake scrollbars
            self.text_committed_scroll_alt.activate(is_dark)
            self.text_working_scroll_alt.activate(is_dark)

            # theme change on Windows can switch sizes; fire a jump to bottom
            self.do_jump_to_bottom_if_necessary()

        # set all the icons
        theme = ""
        oskey = "fbk_"
        rootd = "commit/assets/toolbars/"
        imgfm = ".png"
        if not force_gen:
            force_gen = self.app.config.ReadBool("view/gen_icons")

        # are we dark? if so we want the inverted theme
        if is_dark:
            theme = "inv_"

        # do our platform-specific stuff, but only if the user hasn't chosen to use generic icons
        if not force_gen:
            # are we Mac or Windows?
            if self.app.platform == "mac":
                oskey = "mac_"
            elif self.app.platform == "win":
                oskey = "win_"

            if oskey != "fbk_":
                # handle the windows/mac cases
                for name, obj in self.icons["tb"].items():
                    this_filename = rootd + theme + oskey + "tb_" + name + imgfm
                    obj.LoadFile(
                        this_filename,
                        wx.BITMAP_TYPE_PNG
                    )

            elif self.app.platform in ["lin", "hku"]:
                # on Linux (GTK), we use the GTK-provided system icons

                # reinitialize the ArtProvider (to handle if user changed icon theme)
                # no idea why this is necessary, but the icons don't update otherwise :-/ WTF?
                wx.ArtProvider.Push(
                    wx.ArtProvider()
                )
                wx.ArtProvider.Pop()

                # map our icon codes to GTK standard ones
                gtk_map = {
                    "chevron_up": "go-up",
                    "chevron_down": "go-down",
                    "prefs": "preferences-system",
                    "meta": "document-properties",
                    "fullscreen": "view-fullscreen",
                    "fullscreen_back": "view-restore",
                    "menu": "open-menu-symbolic",
                    "quit": "application-exit",
                }

                # set the GTK toolbar icons (statusbar icons will just drop to fallbacks)
                for name, obj in self.icons["tb"].items():
                    obj.CopyFromIcon(
                        wx.ArtProvider.GetIcon(
                            gtk_map[name],
                            client=wx.ART_TOOLBAR
                        )
                    )

        # fallbacks! icons still invalid/missing will now be set to fbk versions
        # we just skip right to here when the user has chosen to force generic icons
        for name, obj in self.icons["tb"].items():
            if force_gen or not obj.IsOk():
                this_filename = rootd + theme + "fbk_tb_" + name + imgfm
                obj.LoadFile(
                    this_filename,
                    wx.BITMAP_TYPE_PNG
                )

        # saved/unsaved dot
        for name, obj in self.icons["sb_dot"].items():
            this_filename = rootd + theme + "sb_dot_" + name + imgfm
            obj.LoadFile(
                this_filename,
                wx.BITMAP_TYPE_PNG
            )

        # statusbar icons too
        for name, obj in self.icons["sb"].items():
            this_filename = rootd + theme + "fbk_" + "sb_" + name + imgfm
            obj.LoadFile(
                this_filename,
                wx.BITMAP_TYPE_PNG
            )

        # now set whatever icons we ended up with to the toolbar items
        self.tool_commit.SetBitmapLabel(
            self.icons["tb"]["chevron_up"]
        ),
        self.tool_pull.SetBitmapLabel(
            self.icons["tb"]["chevron_down"]
        )
        self.tool_prefs.SetBitmapLabel(
            self.icons["tb"]["prefs"]
        )
        self.tool_meta.SetBitmapLabel(
            self.icons["tb"]["meta"]
        )
        if self.IsFullScreen():
            fs_code = "fullscreen_back"
        else:
            fs_code = "fullscreen"
        self.tool_fullscreen.SetBitmapLabel(
            self.icons["tb"][fs_code]
        )
        self.tool_menu.SetBitmap(
            self.icons["tb"]["menu"]
        )
        self.tool_exit.SetBitmap(
            self.icons["tb"]["quit"]
        )

        # now set the statusbar icons
        if self.file.needs_save:
            dot_code = "unsaved"
        else:
            dot_code = "saved"
        self.sb_unsaved.SetBitmap(
            self.icons["sb_dot"][dot_code]
        )
        self.sb_meta_button.SetBitmap(
            self.icons["sb"]["meta"]
        )
        self.sb_theme_button.SetBitmap(
            self.icons["sb"]["theme"]
        )
        self.sb_ratio_button.SetBitmap(
            self.icons["sb"]["ratio"]
        )
        self.sb_menu_button.SetBitmap(
            self.icons["sb"]["menu"]
        )

        # remember the mode!
        self.is_dark_mode = is_dark

        # force an update
        self.Layout()
        self.Refresh()
        self.Update()
        self.Thaw()

    # preferences have been updated; refresh some things
    def do_prefs_update(self):
        # when prefs are saved, update word count
        self.update_word_count()

        # update font
        self.display_font.SetNativeFontInfo(
            self.app.config.Read("view/font")
        )
        self.text_working_widget.SetFont(self.display_font)
        self.text_committed_widget.SetFont(self.display_font)

        # show/hide the statusbar
        should_show_statusbar = self.app.config.ReadBool("statusbar/show")
        self.statusbar_widget.Show(should_show_statusbar)
        self.statusbar_pop.should_show = not should_show_statusbar

        # if we're in fullscreen mode we need to handle toggling the menu and unsaved icon
        if self.IsFullScreen():
            self.tool_unsaved.Show(not should_show_statusbar)
            self.tool_menu.Show(not should_show_statusbar)

        # we always need to handle the meta toolbar item
        self.tool_meta.Show(not should_show_statusbar)

        # show/hide the word count
        self.sb_panel_2.Show(
            self.app.config.ReadBool("statusbar/show_word_count")
        )

        # show/hide the session word count
        self.sb_panel_3.Show(
            self.app.config.ReadBool("statusbar/show_session_word_count")
        )

        # update the fullscreen ratio
        self.menu_ratio.FindItemByPosition(
            self.app.config.ReadInt("view/fs_ratio")
        ).Check(True)
        self.do_fs_ratio_change(
            self.IsFullScreen()
        )

        # update the theme selector and fire a theme change
        if self.is_full_screen and self.app.config.ReadInt("view/fs_theme") > -1:
            target_setting = "view/fs_theme"
        else:
            target_setting = "view/theme"
        self.menu_theme.FindItemByPosition(
            self.app.config.ReadInt(target_setting)
        ).Check(True)
        self.do_theme_change()

        # make sure the toolbar and statusbar refreshes (sometimes doesn't automatically on Mac??)
        self.toolbar_widget.Layout()
        self.statusbar_widget.Layout()
        self.toolbar_widget.Refresh()
        self.statusbar_widget.Refresh()

    # set the fullscreen ratio
    def do_fs_ratio_change(self, enable=False):
        target_size = wx.Size(-1, -1)
        target_spacer_ratio = 0

        if enable:
            target_ratio = self.app.constants.fs_ratio_map[
                self.app.config.ReadInt("view/fs_ratio")
            ][0]

            # ratio of 0 means don't limit the size at all and let it go fullscreen for real
            if target_ratio != 0:
                this_display_size = wx.GetDisplaySize()
                this_target_width = round(this_display_size[1] * target_ratio)
                this_base_status_offset = round((this_display_size[0] - this_target_width) / 2)
                this_left_status_offset = this_base_status_offset
                this_right_status_offset = this_base_status_offset

                if self.app.platform == "win":
                    this_left_status_offset -= 8
                    this_right_status_offset -= 28

                target_size.SetWidth(this_target_width)
                target_spacer_ratio = 1

        self.main_panel.SetMaxSize(target_size)
        self.fs_spacer_1.SetProportion(target_spacer_ratio)
        self.fs_spacer_2.SetProportion(target_spacer_ratio)

        # keep pinned at the bottom when appropriate
        self.do_jump_to_bottom_if_necessary()

        # redraw/update
        self.GetSizer().Layout()
        self.Refresh()

    # play a sound
    def do_play_sound(self, soundfile):
        if self.app.platform == "lin":
            try:
                os.system("paplay " + str(soundfile.resolve()))
            except:  # noqa - we really just want to fail *silently* (haha)
                pass
        else:
            # sound for other platforms
            this_sound = Sound(str(soundfile.resolve()))
            if this_sound.IsOk():
                this_sound.Play()

    # update the statusbar (field #1)
    def do_update_statusbar(self, status_text, must_match=False):
        # if must_match is set, we only want to update status if the current status matches
        # we use this for the welcome message; don't clear if it has already been changed
        if (
            must_match and
            must_match != self.statusbar_1.GetLabel()
        ):
            return

        # update with the requested text
        self.statusbar_1.SetLabel(status_text)
        self.statusbar_pop.SetLabel(status_text)

    # check for text clipboard
    @staticmethod
    def check_for_text_clipboard():
        # is the clipboard texty?
        if (
            wx.TheClipboard.IsSupported(wx.DataFormat(wx.DF_UNICODETEXT)) or
            wx.TheClipboard.IsSupported(wx.DataFormat(wx.DF_TEXT))
        ):
            return True

        return False

    # toggle fullscreen status
    def do_toggle_fullscreen(self, event=None, force=False):
        self.Show()  # ugh; silly buggy bug bug!
        should_fullscreen = (not self.IsFullScreen() or force)
        has_statusbar = self.app.config.ReadBool("statusbar/show")
        skip_fs_action = False

        if event is not None:
            event.Skip()

            # if we got an event, it means this was triggered by the Mac green button
            # this means we need to get the fullscreen status a different way (reasons??)
            # it also does its own toggle, so we have to skip the toggle step later...sigh
            should_fullscreen = event.IsFullScreen()
            skip_fs_action = True

        if should_fullscreen:
            # set the fullscreen icon to the 'back' style
            target_fs_icon = self.icons["tb"]["fullscreen_back"]
        else:
            # set the fullscreen icon back to the normal
            target_fs_icon = self.icons["tb"]["fullscreen"]

        # apply the fullscreen icon that we picked
        self.tool_fullscreen.SetBitmapLabel(
            target_fs_icon
        )

        # show/hide the toolbar unsaved and menu buttons (only if statusbar is hidden)
        self.tool_unsaved.Show(
            should_fullscreen and not has_statusbar
        )
        self.tool_menu.Show(
            should_fullscreen and not has_statusbar
        )

        # show/hide the exit button
        self.tool_exit.Show(
            should_fullscreen
        )

        # show/hide the statusbar buttons and save indicator
        self.sb_fs_buttons.Show(should_fullscreen)
        self.sb_unsaved.Show(should_fullscreen)

        if not skip_fs_action:
            # do the actual fullscreen toggle, but only if we don't have the skip flag.
            # skip flag on Mac green button, since the toggle is already happening
            self.ShowFullScreen(
                should_fullscreen,
                wx.FULLSCREEN_NOMENUBAR | wx.FULLSCREEN_NOTOOLBAR | wx.FULLSCREEN_NOBORDER | wx.FULLSCREEN_NOCAPTION
            )

        # set the window flag
        self.is_full_screen = should_fullscreen

        # handle the ratio change
        self.do_fs_ratio_change(should_fullscreen)

        # save fullscreen status
        self.app.config.WriteBool("window/is_fullscreen", should_fullscreen)

        # update the theme selector
        if should_fullscreen and self.app.config.ReadInt("view/fs_theme") > -1:
            target_setting = "view/fs_theme"
        else:
            target_setting = "view/theme"
        self.menu_theme.FindItemByPosition(
            self.app.config.ReadInt(target_setting)
        ).Check(True)

        # handle the theme change
        self.do_theme_change(
            full_screen=should_fullscreen
        )

    # update the title and saved indicators
    def update_title(self):
        old_title = self.GetTitle()
        new_title = self.app.constants.app_display_name

        if self.file.path.is_file():
            # prepend the file name if we have one
            new_title = self.file.path.name + " – " + new_title

            # set the icon (only does it on Mac)
            self.SetRepresentedFilename(
                str(self.file.path.resolve())
            )
        else:
            self.SetRepresentedFilename("")

        if self.file.needs_save:
            # handle the unsaved indicators
            if self.app.platform != "mac":
                # only put the dot in the title on non-Mac platforms
                new_title = "• " + new_title

            # fullscreen statusbar icon to unsaved
            sb_unsaved_bitmap = self.icons["sb_dot"]["unsaved"]
        else:
            # fullscreen statusbar icon to saved
            sb_unsaved_bitmap = self.icons["sb_dot"]["saved"]

        if old_title != new_title:
            # only if the title has actually changed, go!

            # set the saved/unsaved bitmap in the statusbar (only shown in fullscreen mode)
            self.tool_unsaved.SetBitmap(sb_unsaved_bitmap)
            self.tool_unsaved.Refresh()
            self.sb_unsaved.SetBitmap(sb_unsaved_bitmap)
            self.sb_unsaved.Refresh()

            # set the window title
            self.SetTitle(new_title)

            # set the Mac modified doojobber in the window close button
            self.OSXSetModified(self.file.needs_save)

    # calculate and show the word count
    def update_word_count(self, reset_doc_count=False):
        this_length = len(self.text_committed_widget.GetValue().split())

        if reset_doc_count:
            self.word_count_start = this_length

        if self.app.config.ReadBool("statusbar/show_word_count"):
            format_length = locale.format_string("%d", this_length, grouping=True, monetary=True)

            # insert into statusbar
            if this_length == 1:
                self.statusbar_2.SetLabel(format_length + " " + self._("committed word"))
            else:
                self.statusbar_2.SetLabel(format_length + " " + self._("committed words"))
        else:
            self.statusbar_2.SetLabel("")

        if self.app.config.ReadBool("statusbar/show_session_word_count"):
            session_words = this_length - self.word_count_start
            if session_words < 0:
                session_words = 0

            format_length = locale.format_string("%d", session_words, grouping=True)

            # insert into statusbar
            if session_words == 1:
                self.statusbar_3.SetLabel(format_length + " " + self._("new word in this session"))
            else:
                self.statusbar_3.SetLabel(format_length + " " + self._("new words in this session"))
        else:
            self.statusbar_3.SetLabel("")

    @staticmethod
    def gtk_skip(event):
        if wx.Platform == "__WXGTK__":
            event.Skip()

    # when the window opens
    def event_window_show(self, event):
        self.Freeze()

        # tweak the margins in GTK, which, for some reason, can't be done directly from wxPython (!?)
        # this has to be done after the window is shown so all the handles are right
        if not self.init_completed and wx.Platform == "__WXGTK__":
            # get GTK ids for the wxPython text fields...but these are actually GtkScrolledWindows!
            # we'll have to search *their* children to find the GtkTextView objects
            committed_gtk_id = hex(self.text_committed_widget.GetGtkWidget().__int__())
            working_gtk_id = hex(self.text_working_widget.GetGtkWidget().__int__())

            def gtk_recurse(this_obj, gtk_catch_for=None):
                # if we have children, recurse!
                if (
                    hasattr(this_obj, "get_children") and
                    callable(this_obj.get_children)
                ):
                    # if the current object matches the GtkScrolledWindow
                    # set the flag so we can grab the GtkTextView inside it
                    if committed_gtk_id in str(this_obj):
                        gtk_catch_for = "committed"
                    elif working_gtk_id in str(this_obj):
                        gtk_catch_for = "working"

                    for sub_obj in this_obj.get_children():
                        gtk_recurse(sub_obj, gtk_catch_for=gtk_catch_for)

                # we found the text fields we're looking for; save them for later reference
                if this_obj.get_name() == "GtkTextView":
                    if gtk_catch_for == "committed":
                        self.gtk_text_committed_widget = this_obj
                    elif gtk_catch_for == "working":
                        self.gtk_text_working_widget = this_obj

            def gtk_find_window():
                # this is the hex handle to our window
                gtk_id_hex = hex(
                    self.GetGtkWidget().__int__()
                )

                # we have to loop through and find the text fields
                for window in Gtk.Window.list_toplevels():
                    if gtk_id_hex in str(window):
                        gtk_recurse(window)

                # now that we have the text fields, set their margins with GTK methods
                self.gtk_text_committed_widget.set_left_margin(10)
                self.gtk_text_committed_widget.set_right_margin(10)
                self.gtk_text_working_widget.set_left_margin(10)
                self.gtk_text_working_widget.set_right_margin(10)

                # jump to bottom (GTK needs it again?)
                self.do_jump_to_bottom()

            wx.CallAfter(gtk_find_window)

        # set focus to working widget
        self.text_working_widget.SetFocus()

        # the window has fully initialized
        self.init_completed = True

        event.Skip()
        self.Thaw()

    # when the app closesd
    def event_window_close(self, event):
        if (
            self.file.handle_unsaved_changes() and
            self.file.lockfile.is_file()
        ):
            self.file.lockfile.unlink()

        if not self.app.is_config_reset:
            self.app.recents.update(
                closing_file=str(
                    self.file.path.resolve()
                )
            )

        # set a flag so the unfocus event knows we're closing
        self.is_closing = True

        event.Skip()

    def event_window_resize(self, event=None):
        # always do the normal action
        if event is not None:
            event.Skip()

        """
        Event Codes:
            10086 = size
            10088 = move/restore
            10102 = maximize
            (but there are others, facepalm)
        """

        # do a jump to bottom, but only if conditions warrant is
        self.do_jump_to_bottom_if_necessary()

        # config object
        config = self.app.config

        # get size and position
        window_size = self.GetSize().Get()
        window_pos = self.GetPosition().Get()

        # now save it all back to the prefs
        config.WriteBool("window/is_maximized", self.IsMaximized())
        config.WriteBool("window/is_fullscreen", self.IsFullScreen())

        # but don't write the size and pos
        if not self.IsMaximized() and not self.IsFullScreen():
            config.WriteInt("window/size_x", window_size[0])
            config.WriteInt("window/size_y", window_size[1])
            config.WriteInt("window/pos_x", window_pos[0])
            config.WriteInt("window/pos_y", window_pos[1])

    # intercept key presses (can do special shortcuts and stuff here)
    def event_window_keys(self, event):
        event.Skip()

        if not self.app.constants.debug:
            if self.app.platform == "mac":
                enable_debug = (
                    event.ControlDown() and
                    event.ShiftDown() and
                    event.GetKeyCode() == 68
                )
            else:
                enable_debug = (
                    event.ControlDown() and
                    event.AltDown() and
                    event.GetKeyCode() == 68
                )

            # enable debug menu on CTRL-ALT-D (if not already enabled)
            if enable_debug:
                self.menu_main.Append(
                    self.menu_debug,
                    self.menu_debug.GetTitle()
                )

                if self.app.platform == "win":
                    self.menu_main_alt.append(
                        self.menu_debug,
                        self.menu_debug.GetTitle()
                    )

                self.app.constants.debug = True

    # when you do mouse stuff in either of the text fields
    def event_text_field_mouse_click(self, event):
        # do the default action
        event.Skip()

        # refresh the edit menu (enable/disable)
        # force disable if we're not the working widget
        self.do_edit_menu_refresh(
            all_disabled=(event.GetEventObject().GetName() != "text_working_widget")
        )

    def event_committed_key_press(self, event):
        if (
            self.app.config.ReadBool("options/disable-scrolling") and
            event.GetKeyCode() in [
                wx.WXK_PAGEDOWN,
                wx.WXK_PAGEUP,
                wx.WXK_UP,
                wx.WXK_DOWN,
                wx.WXK_LEFT,
                wx.WXK_RIGHT,
                wx.WXK_HOME,
                wx.WXK_END
            ]
        ):
            event.StopPropagation()
        elif event.GetKeyCode() == wx.WXK_END:
            self.do_jump_to_bottom()
        elif event.GetKeyCode() == wx.WXK_HOME:
            self.text_committed_widget.ShowPosition(0)
        else:
            event.Skip()

        # update the edit menu; force to false because working isn't focused
        self.do_edit_menu_refresh(all_disabled=True)

    # when we scroll in the committed window, we sometimes want to stop it (there's a setting!)
    def event_committed_scroll(self, event):
        if not self.scroll_bypass and self.app.config.ReadBool("options/disable-scrolling"):
            event.StopPropagation()
        else:
            event.Skip()

    # fires on any keydown event inside the editing window
    def event_working_key_press(self, event):
        current_insertion_point = event.GetEventObject().GetInsertionPoint()

        # refresh the edit menu (enable/disable)
        self.do_edit_menu_refresh()

        # detect various keyboard shortcuts
        if (
                event.GetModifiers() != wx.MOD_SHIFT and
                event.GetKeyCode() in [wx.WXK_RETURN, wx.WXK_NUMPAD_ENTER]
        ):
            # commit unless we're holding the shift key
            self.do_commit(True)
        elif (
                event.GetKeyCode() == wx.WXK_BACK and
                current_insertion_point == 0 and
                self.text_working_widget.GetStringSelection() == ""
        ):
            # do a pull
            self.do_pull()
        else:
            # in any other case, keypress should do whatever it normally does
            event.Skip()

    # whenever the text is changed
    def event_working_text_changed(self, event):
        # do the default action
        event.Skip()

        if not self.ignore_edits:
            # mark that the file needs saving
            self.file.needs_save = True

            # update the document title to reflect unsaved status
            self.update_title()

    # when something is pasted
    def event_working_paste(self, _event):
        # only allow pasting plain text!
        clipboard = wx.Clipboard()

        # open clipboard
        if (
                self.check_for_text_clipboard() and
                clipboard.Open()
        ):
            # get clipboard data
            data = wx.TextDataObject()
            clipboard.GetData(data)
            clipboard.Close()

            # insert text into field
            self.text_working_widget.WriteText(
                data.GetText()
            )
        else:
            # nothing valid to paste
            wx.Bell()

    def event_working_focus(self, event):
        event.Skip()

        # refresh the edit menu (enable/disable)
        self.do_edit_menu_refresh()

    def event_working_focus_lost(self, event):
        event.Skip()

        # refresh the edit menu (all disabled; lost focus)
        # but NOT if the window is in the process of closing!
        if not self.is_closing:
            self.do_edit_menu_refresh(all_disabled=True)

    # window statusbar handler (replicates behavior of built-in statusbar widget)
    def event_statusbar_update(self, event):
        event.Skip()
        status_string = ""
        this_event_type = event.GetEventType()

        # now go through the actual statusbar stuff
        if this_event_type == wx.wxEVT_MENU_HIGHLIGHT:
            this_menu = event.GetMenu()
            if this_menu is not None:
                this_menu_item = this_menu.FindItemById(
                    event.GetMenuId()
                )
                if this_menu_item is not None:
                    status_string = this_menu_item.GetHelp()

        elif this_event_type == wx.wxEVT_ENTER_WINDOW:
            this_obj_id = event.GetId()

            # toolbar items
            if this_obj_id == self.tool_commit.GetId():
                status_string = self._("Commit working text")

            if this_obj_id == self.tool_pull.GetId():
                status_string = self._("Pull back last committed paragraph")

            if this_obj_id == self.tool_prefs.GetId():
                status_string = self._("Open preferences dialog")

            if this_obj_id == self.tool_fullscreen.GetId():
                status_string = self._("Toggle full screen mode")

            if this_obj_id == self.tool_exit.GetId():
                status_string = self._("Quit this program")

            # statusbar items (some can be in either the toolbar or statusbar)
            if (
                this_obj_id == self.tool_unsaved.GetId() or
                this_obj_id == self.sb_unsaved.GetId()
            ):
                if self.file.needs_save:
                    status_string = self._("You have unsaved changes (click to save)")
                else:
                    status_string = self._("You have no unsaved changes")

            if (
                this_obj_id == self.tool_meta.GetId() or
                this_obj_id == self.sb_meta_button.GetId()
            ):
                status_string = self._("Edit document metadata")
            if this_obj_id == self.sb_theme_button.GetId():
                status_string = self._("Change display theme")
            if(
                this_obj_id == self.tool_menu.GetId() or
                this_obj_id == self.sb_menu_button.GetId()
            ):
                status_string = self._("Open the main menu")

            if this_obj_id == self.sb_ratio_button.GetId():
                status_string = self._("Adjust editor display ratio")

        self.do_update_statusbar(status_string)

    # theme menu button
    def event_statusbar_button_theme(self, event):
        event.GetEventObject().PopupMenu(
            self.menu_theme,
            wx.UP,
            event=event
        )
        self.gtk_skip(event)

    # selection of a theme
    def event_statusbar_theme_select(self, event):
        this_menu = event.GetEventObject()
        this_selection = event.GetId()

        # if we're in full-screen AND FS is not set to follow base, then we want to change full-screen theme
        if self.is_full_screen and self.app.config.ReadInt("view/fs_theme") > -1:
            to_change = "view/fs_theme"
        else:
            to_change = "view/theme"

        # handle adjustment of the theme selector
        self.app.config.WriteInt(
            to_change,
            this_menu.FindChildItem(this_selection)[1]
        )

        self.do_prefs_update()

    # statusbar menu button
    def event_statusbar_button_menu(self, event):
        event.GetEventObject().PopupMenu(
            self.menu_alt,
            wx.UP,
            event=event
        )
        self.gtk_skip(event)

    # statusbar ratio button
    def event_statusbar_button_ratio(self, event):
        event.GetEventObject().PopupMenu(
            self.menu_ratio,
            wx.UP,
            event=event
        )
        self.gtk_skip(event)

    # selection of a ratio
    def event_statusbar_ratio_select(self, event):
        this_menu = event.GetEventObject()
        this_selection = event.GetId()

        # handle adjustment of the fullscreen ratio selector
        self.app.config.WriteInt(
            "view/fs_ratio",
            this_menu.FindChildItem(this_selection)[1]
        )

        self.do_prefs_update()

    def event_menu_file_new(self, event):
        self.file.new()

    def event_menu_file_open(self, event):
        self.file.open()

    def event_menu_file_save(self, event):
        self.file.save()

    def event_menu_file_save_as(self, event):
        self.file.save(save_as=True)

    # open recent item from menu
    def event_menu_file_recent_open(self, event):
        real_id = event.GetId() - 10000
        self.file.open(
            self.app.recents.file_list[real_id]
        )

    # reset recent items menu
    def event_menu_file_recent_reset(self, event):  # noqa
        self.app.recents.reset(self)

    def event_menu_file_prefs(self, event):
        # pop open the "prefs" window
        prefs_win = PrefsWin(self)
        pref_result = prefs_win.ShowModal()

        if pref_result == wx.ID_SAVE:
            self.do_prefs_update()

        self.gtk_skip(event)

    def event_menu_file_import(self, event):
        if not self.file.handle_unsaved_changes():
            return

        # hand off to the importer
        self.convert.do_import_file(self)

    def event_menu_file_export(self, event):
        if not self.file.handle_unsaved_changes():
            return

        # hand off to the exporter
        self.convert.do_export_file(self)

    def event_menu_file_export_cb(self, event):
        # hand off to the exporter
        self.convert.do_export_clipboard(self)

    def event_menu_file_print(self, event):
        self.file.print(self.convert)

    def event_menu_file_exit(self, event):
        self.gtk_skip(event)
        self.Close()
        self.Destroy()

    def event_menu_edit_undo(self, event):
        event.Skip()

        # perform an undo (only in the working area)
        self.text_working_widget.Undo()

    def event_menu_edit_cut(self, event):
        # Cut from the focused area (if allowed)
        self.FindFocus().Cut()

    def event_menu_edit_copy(self, event):
        # Copy from the focused area (if allowed)
        self.FindFocus().Copy()

    def event_menu_edit_paste(self, event):
        # Paste into the focused area (if allowed)
        self.FindFocus().Paste()

    def event_menu_edit_delete(self, event):
        # Delete the selected text in the working area (if any)
        my_selection = self.text_working_widget.GetSelection()
        self.text_working_widget.Remove(my_selection[0], my_selection[1])

    def event_menu_select_all(self, event):
        # Select all text in the focused area
        self.FindFocus().SelectAll()

    def event_menu_edit_commit(self, event):
        self.do_commit()
        self.gtk_skip(event)

    def event_menu_edit_pull(self, event):
        self.do_pull()
        self.gtk_skip(event)

    def event_menu_edit_find(self, event):
        SearchFindDialog(self)

    def event_menu_view_newwin(self, event):
        MainWin(None, pathlib.Path(), True).Show()

    def event_menu_view_meta(self, event):
        # pop open the metadata window
        if not self.meta_open:
            meta_window = MetaWin(self)
            meta_window.SetSize(
                self.FromDIP(
                    wx.Size(500, -1)
                )
            )
            meta_window.Center()
            meta_window.Show()

    def event_menu_view_minimize(self, event):
        # Minimize the window (which is called "iconize" for some reason)
        self.Iconize(True)

    def event_menu_view_maximize(self, event):
        _ = self._

        # Toggle maximize status
        self.Maximize(
            not self.IsMaximized()
        )

        # what should the menu label be?
        if self.IsMaximized():
            menu_label = _("Restore")
        else:
            menu_label = _("Maximize")

        # set menu label
        self.menu_view_maximize.SetItemLabel(menu_label)

    def event_menu_view_fullscreen(self, event):
        self.do_toggle_fullscreen()
        self.gtk_skip(event)

    def event_menu_help_website(self, event):
        # open the app website
        webbrowser.open(
            self.app.constants.app_website,
            True,
            True
        )

    def event_menu_help_manual(self, event):
        # open the help/manual website
        webbrowser.open(
            self.app.constants.app_help_url,
            True,
            True
        )

    def event_menu_help_donate(self, event):
        # open the donation website
        webbrowser.open(
            self.app.constants.app_donate_url,
            True,
            True
        )

    def event_menu_help_updates(self, event):
        # fire an update
        self.app.updater.do_update_check(self)

    def event_menu_help_pandoc(self, event):
        # check for pandoc installation
        self.app.config.WriteBool(
            "params/pandoc_check",
            True
        )
        self.convert.pandoc_check(self)

    def event_menu_help_about(self, event):
        # pop open the "about" window
        about_win = AboutWin(self)
        about_win.ShowWindowModal()

    def event_menu_debug(self, event):
        this_id = event.GetId()

        if this_id == self.menu_debug_wxinsp.GetId():
            self.app.enable_debug_mode(wxinsp=True)
        elif this_id == self.menu_debug_stdio.GetId():
            self.app.enable_debug_mode(st=True)
        elif this_id == self.menu_debug_fault.GetId():
            self.app.enable_debug_mode(fh=True)
        elif this_id == self.menu_debug_all.GetId():
            self.app.enable_debug_mode(
                wxinsp=True,
                st=True,
                fh=True
            )
