import locale       # Python Locale - https://docs.python.org/3/library/locale.html
import wx           # wxPython - https://wxpython.org/

from commit.ui.commit_ui_design import MetaWin_Base   # local - about window design from wxFormBuilder
from commit.includes.customobjects import WindowThemeable   # local - custom objects


class MetaWin(MetaWin_Base, WindowThemeable):
    def __init__(self, parent):
        # call parent inits
        MetaWin_Base.__init__(self, parent)
        WindowThemeable.__init__(self, parent)

        self.meta = parent.file.meta
        self.parent = parent

        self.is_changed = False

        self.meta_title.SetValue(
            self.meta["title"]
        )
        self.meta_title.SetFocus()
        self.meta_author.SetValue(
            self.meta["author"][0]
        )
        self.meta_subject.SetValue(
            self.meta["subject"]
        )
        self.meta_description.SetValue(
            self.meta["description"]
        )
        self.meta_keywords.SetValue(
            ", ".join(self.meta["keywords"])
        )

        doc_date = wx.DateTime()
        doc_date.ParseISOCombined(self.meta["date"])

        self.meta_doc_datetime.SetValue(
            doc_date.Format(
                self.meta_doc_datetime.datetime_fmt
            )
        )

        save_date = wx.DateTime.FromTimeT(0)
        if self.meta["last_save"] is not None:
            save_date.ParseISOCombined(self.meta["last_save"])

        self.meta_save_datetime.SetValue(
            save_date.Format(
                self.meta_save_datetime.datetime_fmt
            )
        )
        # default colors for disabled don't look right; force to theme colors
        self.meta_save_datetime.SetForegroundColour(self.theme_colors["disabled_text"])
        self.meta_save_datetime.SetBackgroundColour(self.theme_colors["disabled"])

        # get the full word count (can be more than the "committed" count shown in the statusbar
        this_length =\
            len(self.parent.text_committed_widget.GetValue().split()) +\
            len(self.parent.text_working_widget.GetValue().split())
        self.meta_show_word_count.SetLabel(
            locale.format_string(
                "%d",
                this_length,
                grouping=True,
                monetary=True
            )
        )

        # get the file size
        if self.parent.file.path.is_file():
            size = self.parent.file.path.stat().st_size / 1024
            unit = "KB"

            if size > 1024:
                size = size / 1024
                unit = "MB"

            file_size = locale.format_string(
                "%.2f " + unit,
                size,
                grouping=True,
                monetary=True
            )
        else:
            file_size = "N/A (unsaved)"
        self.meta_show_file_size.SetLabel(file_size)

        # output the generator (with "Commit" removed; i.e. just the version)
        self.meta_show_generator.SetLabel(
            self.meta["generator"].replace(
                self.parent.app.constants.app_display_name + " ", ""
            )
        )

        # show the CTXT format version from the metadata
        self.meta_show_ctxt_version.SetLabel(
            self.meta["ctxt_ver"]
        )

        # show/hide
        self.Bind(
            wx.EVT_SHOW,
            self.event_show_hide
        )

    def event_show_hide(self, event):
        event.Skip()
        self.parent.meta_open = event.IsShown()

    def do_meta_save(self, event):
        event.Skip()

        if self.is_changed:
            # if there was a change

            # is the date entry valid?
            if not self.meta_doc_datetime.IsValid():
                wx.MessageBox(
                    self._("Invalid document date."),
                    self._("Error"),
                    wx.OK | wx.ICON_ERROR,
                    self
                )
                return False

            # update metadata
            self.meta["title"] = self.meta_title.GetValue()
            self.meta["author"][0] = self.meta_author.GetValue()
            self.meta["subject"] = self.meta_subject.GetValue()
            self.meta["description"] = self.meta_description.GetValue()
            self.meta["keywords"] = [
                x.strip() for x in self.meta_keywords.GetValue().split(",")
            ]

            doc_datetime = wx.DateTime()
            doc_datetime.ParseFormat(
                self.meta_doc_datetime.GetValue(),
                self.meta_doc_datetime.datetime_fmt
            )

            self.meta["date"] = doc_datetime.FormatISOCombined()

            # set file as unsaved and update window title
            self.parent.file.needs_save = True
            self.parent.update_title()

        # close the prefs window
        self.Close()

    def set_updated(self, event):
        event.Skip()

        self.is_changed = True
